#   Copyright 2013 Nebula Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from unittest import mock
from unittest.mock import call

from osc_lib import exceptions
from osc_lib import utils

from openstackclient.identity import common
from openstackclient.identity.v3 import project
from openstackclient.tests.unit.identity.v3 import fakes as identity_fakes


class TestProject(identity_fakes.TestIdentityv3):
    def setUp(self):
        super().setUp()

        # Get a shortcut to the DomainManager Mock
        self.domains_mock = self.identity_client.domains
        self.domains_mock.reset_mock()

        # Get a shortcut to the ProjectManager Mock
        self.projects_mock = self.identity_client.projects
        self.projects_mock.reset_mock()


class TestProjectCreate(TestProject):
    domain = identity_fakes.FakeDomain.create_one_domain()

    columns = (
        'description',
        'domain_id',
        'enabled',
        'id',
        'is_domain',
        'name',
        'parent_id',
        'tags',
    )

    def setUp(self):
        super().setUp()

        self.project = identity_fakes.FakeProject.create_one_project(
            attrs={'domain_id': self.domain.id}
        )
        self.domains_mock.get.return_value = self.domain
        self.projects_mock.create.return_value = self.project
        self.datalist = (
            self.project.description,
            self.project.domain_id,
            True,
            self.project.id,
            False,
            self.project.name,
            self.project.parent_id,
            self.project.tags,
        )
        # Get the command object to test
        self.cmd = project.CreateProject(self.app, None)

    def test_project_create_no_options(self):
        arglist = [
            self.project.name,
        ]
        verifylist = [
            ('parent', None),
            ('enabled', True),
            ('name', self.project.name),
            ('tags', []),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': self.project.name,
            'domain': None,
            'description': None,
            'enabled': True,
            'parent': None,
            'tags': [],
            'options': {},
        }
        # ProjectManager.create(name=, domain=, description=,
        #                       enabled=, **kwargs)
        self.projects_mock.create.assert_called_with(**kwargs)

        collist = (
            'description',
            'domain_id',
            'enabled',
            'id',
            'is_domain',
            'name',
            'parent_id',
            'tags',
        )
        self.assertEqual(collist, columns)
        datalist = (
            self.project.description,
            self.project.domain_id,
            True,
            self.project.id,
            False,
            self.project.name,
            self.project.parent_id,
            self.project.tags,
        )
        self.assertEqual(datalist, data)

    def test_project_create_description(self):
        arglist = [
            '--description',
            'new desc',
            self.project.name,
        ]
        verifylist = [
            ('description', 'new desc'),
            ('enabled', True),
            ('name', self.project.name),
            ('parent', None),
            ('tags', []),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': self.project.name,
            'domain': None,
            'description': 'new desc',
            'enabled': True,
            'parent': None,
            'tags': [],
            'options': {},
        }
        # ProjectManager.create(name=, domain=, description=,
        #                       enabled=, **kwargs)
        self.projects_mock.create.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_project_create_domain(self):
        arglist = [
            '--domain',
            self.project.domain_id,
            self.project.name,
        ]
        verifylist = [
            ('domain', self.project.domain_id),
            ('enabled', True),
            ('name', self.project.name),
            ('parent', None),
            ('tags', []),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': self.project.name,
            'domain': self.project.domain_id,
            'description': None,
            'enabled': True,
            'parent': None,
            'tags': [],
            'options': {},
        }
        # ProjectManager.create(name=, domain=, description=,
        #                       enabled=, **kwargs)
        self.projects_mock.create.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_project_create_domain_no_perms(self):
        arglist = [
            '--domain',
            self.project.domain_id,
            self.project.name,
        ]
        verifylist = [
            ('domain', self.project.domain_id),
            ('enabled', True),
            ('name', self.project.name),
            ('parent', None),
            ('tags', []),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        mocker = mock.Mock()
        mocker.return_value = None

        with mock.patch("osc_lib.utils.find_resource", mocker):
            columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': self.project.name,
            'domain': self.project.domain_id,
            'description': None,
            'enabled': True,
            'parent': None,
            'tags': [],
            'options': {},
        }
        self.projects_mock.create.assert_called_with(**kwargs)
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_project_create_enable(self):
        arglist = [
            '--enable',
            self.project.name,
        ]
        verifylist = [
            ('enabled', True),
            ('name', self.project.name),
            ('parent', None),
            ('tags', []),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': self.project.name,
            'domain': None,
            'description': None,
            'enabled': True,
            'parent': None,
            'tags': [],
            'options': {},
        }
        # ProjectManager.create(name=, domain=, description=,
        #                       enabled=, **kwargs)
        self.projects_mock.create.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_project_create_disable(self):
        arglist = [
            '--disable',
            self.project.name,
        ]
        verifylist = [
            ('enabled', False),
            ('name', self.project.name),
            ('parent', None),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': self.project.name,
            'domain': None,
            'description': None,
            'enabled': False,
            'parent': None,
            'tags': [],
            'options': {},
        }
        # ProjectManager.create(name=, domain=,
        #                       description=, enabled=, **kwargs)
        self.projects_mock.create.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_project_create_property(self):
        arglist = [
            '--property',
            'fee=fi',
            '--property',
            'fo=fum',
            self.project.name,
        ]
        verifylist = [
            ('properties', {'fee': 'fi', 'fo': 'fum'}),
            ('name', self.project.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': self.project.name,
            'domain': None,
            'description': None,
            'enabled': True,
            'parent': None,
            'fee': 'fi',
            'fo': 'fum',
            'tags': [],
            'options': {},
        }
        # ProjectManager.create(name=, domain=, description=,
        #                       enabled=, **kwargs)
        self.projects_mock.create.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_project_create_is_domain_false_property(self):
        arglist = [
            '--property',
            'is_domain=false',
            self.project.name,
        ]
        verifylist = [
            ('parent', None),
            ('enabled', True),
            ('name', self.project.name),
            ('tags', []),
            ('properties', {'is_domain': 'false'}),
            ('name', self.project.name),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': self.project.name,
            'domain': None,
            'description': None,
            'enabled': True,
            'parent': None,
            'is_domain': False,
            'tags': [],
            'options': {},
        }
        self.projects_mock.create.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_project_create_is_domain_true_property(self):
        arglist = [
            '--property',
            'is_domain=true',
            self.project.name,
        ]
        verifylist = [
            ('parent', None),
            ('enabled', True),
            ('name', self.project.name),
            ('tags', []),
            ('properties', {'is_domain': 'true'}),
            ('name', self.project.name),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': self.project.name,
            'domain': None,
            'description': None,
            'enabled': True,
            'parent': None,
            'is_domain': True,
            'tags': [],
            'options': {},
        }
        self.projects_mock.create.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_project_create_is_domain_none_property(self):
        arglist = [
            '--property',
            'is_domain=none',
            self.project.name,
        ]
        verifylist = [
            ('parent', None),
            ('enabled', True),
            ('name', self.project.name),
            ('tags', []),
            ('properties', {'is_domain': 'none'}),
            ('name', self.project.name),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': self.project.name,
            'domain': None,
            'description': None,
            'enabled': True,
            'parent': None,
            'is_domain': None,
            'tags': [],
            'options': {},
        }
        self.projects_mock.create.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_project_create_parent(self):
        self.parent = identity_fakes.FakeProject.create_one_project()
        self.project = identity_fakes.FakeProject.create_one_project(
            attrs={'domain_id': self.domain.id, 'parent_id': self.parent.id}
        )
        self.projects_mock.get.return_value = self.parent
        self.projects_mock.create.return_value = self.project

        arglist = [
            '--domain',
            self.project.domain_id,
            '--parent',
            self.parent.name,
            self.project.name,
        ]
        verifylist = [
            ('domain', self.project.domain_id),
            ('parent', self.parent.name),
            ('enabled', True),
            ('name', self.project.name),
            ('tags', []),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        kwargs = {
            'name': self.project.name,
            'domain': self.project.domain_id,
            'parent': self.parent.id,
            'description': None,
            'enabled': True,
            'tags': [],
            'options': {},
        }

        self.projects_mock.create.assert_called_with(**kwargs)

        collist = (
            'description',
            'domain_id',
            'enabled',
            'id',
            'is_domain',
            'name',
            'parent_id',
            'tags',
        )
        self.assertEqual(columns, collist)
        datalist = (
            self.project.description,
            self.project.domain_id,
            self.project.enabled,
            self.project.id,
            self.project.is_domain,
            self.project.name,
            self.parent.id,
            self.project.tags,
        )
        self.assertEqual(data, datalist)

    def test_project_create_invalid_parent(self):
        self.projects_mock.resource_class.__name__ = 'Project'
        self.projects_mock.get.side_effect = exceptions.NotFound(
            'Invalid parent'
        )
        self.projects_mock.find.side_effect = exceptions.NotFound(
            'Invalid parent'
        )

        arglist = [
            '--domain',
            self.project.domain_id,
            '--parent',
            'invalid',
            self.project.name,
        ]
        verifylist = [
            ('domain', self.project.domain_id),
            ('parent', 'invalid'),
            ('enabled', True),
            ('name', self.project.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaises(
            exceptions.CommandError,
            self.cmd.take_action,
            parsed_args,
        )

    def test_project_create_with_tags(self):
        arglist = [
            '--domain',
            self.project.domain_id,
            '--tag',
            'foo',
            self.project.name,
        ]
        verifylist = [
            ('domain', self.project.domain_id),
            ('enabled', True),
            ('name', self.project.name),
            ('parent', None),
            ('tags', ['foo']),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': self.project.name,
            'domain': self.project.domain_id,
            'description': None,
            'enabled': True,
            'parent': None,
            'tags': ['foo'],
            'options': {},
        }
        self.projects_mock.create.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_project_create_with_immutable_option(self):
        arglist = [
            '--immutable',
            self.project.name,
        ]
        verifylist = [
            ('immutable', True),
            ('description', None),
            ('enabled', True),
            ('name', self.project.name),
            ('parent', None),
            ('tags', []),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': self.project.name,
            'domain': None,
            'description': None,
            'enabled': True,
            'parent': None,
            'tags': [],
            'options': {'immutable': True},
        }
        # ProjectManager.create(name=, domain=, description=,
        #                       enabled=, **kwargs)
        self.projects_mock.create.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_project_create_with_no_immutable_option(self):
        arglist = [
            '--no-immutable',
            self.project.name,
        ]
        verifylist = [
            ('immutable', False),
            ('description', None),
            ('enabled', True),
            ('name', self.project.name),
            ('parent', None),
            ('tags', []),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': self.project.name,
            'domain': None,
            'description': None,
            'enabled': True,
            'parent': None,
            'tags': [],
            'options': {'immutable': False},
        }
        # ProjectManager.create(name=, domain=, description=,
        #                       enabled=, **kwargs)
        self.projects_mock.create.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)


class TestProjectDelete(TestProject):
    project = identity_fakes.FakeProject.create_one_project()

    def setUp(self):
        super().setUp()

        # This is the return value for utils.find_resource()
        self.projects_mock.get.return_value = self.project
        self.projects_mock.delete.return_value = None

        # Get the command object to test
        self.cmd = project.DeleteProject(self.app, None)

    def test_project_delete_no_options(self):
        arglist = [
            self.project.id,
        ]
        verifylist = [
            ('projects', [self.project.id]),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.projects_mock.delete.assert_called_with(
            self.project.id,
        )
        self.assertIsNone(result)

    @mock.patch.object(utils, 'find_resource')
    def test_delete_multi_projects_with_exception(self, find_mock):
        find_mock.side_effect = [self.project, exceptions.CommandError]
        arglist = [
            self.project.id,
            'unexist_project',
        ]
        verifylist = [
            ('projects', arglist),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        try:
            self.cmd.take_action(parsed_args)
            self.fail('CommandError should be raised.')
        except exceptions.CommandError as e:
            self.assertEqual('1 of 2 projects failed to delete.', str(e))

        find_mock.assert_any_call(self.projects_mock, self.project.id)
        find_mock.assert_any_call(self.projects_mock, 'unexist_project')

        self.assertEqual(2, find_mock.call_count)
        self.projects_mock.delete.assert_called_once_with(self.project.id)


class TestProjectList(TestProject):
    domain = identity_fakes.FakeDomain.create_one_domain()
    project = identity_fakes.FakeProject.create_one_project(
        attrs={'domain_id': domain.id}
    )
    projects = identity_fakes.FakeProject.create_projects()

    columns = (
        'ID',
        'Name',
    )
    datalist = (
        (
            project.id,
            project.name,
        ),
    )
    datalists = (
        (
            projects[0].description,
            True,
            projects[0].id,
            projects[0].name,
        ),
        (
            projects[1].description,
            True,
            projects[1].id,
            projects[1].name,
        ),
    )

    def setUp(self):
        super().setUp()

        self.projects_mock.list.return_value = [self.project]

        # Get the command object to test
        self.cmd = project.ListProject(self.app, None)

    def test_project_list_no_options(self):
        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)
        self.projects_mock.list.assert_called_with()

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, tuple(data))

    def test_project_list_long(self):
        arglist = [
            '--long',
        ]
        verifylist = [
            ('long', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)
        self.projects_mock.list.assert_called_with()

        collist = ('ID', 'Name', 'Domain ID', 'Description', 'Enabled')
        self.assertEqual(collist, columns)
        datalist = (
            (
                self.project.id,
                self.project.name,
                self.project.domain_id,
                self.project.description,
                True,
            ),
        )
        self.assertEqual(datalist, tuple(data))

    def test_project_list_domain(self):
        arglist = [
            '--domain',
            self.project.domain_id,
        ]
        verifylist = [
            ('domain', self.project.domain_id),
        ]

        self.domains_mock.get.return_value = self.domain

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)
        self.projects_mock.list.assert_called_with(
            domain=self.project.domain_id
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, tuple(data))

    def test_project_list_domain_no_perms(self):
        arglist = [
            '--domain',
            self.project.domain_id,
        ]
        verifylist = [
            ('domain', self.project.domain_id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        mocker = mock.Mock()
        mocker.return_value = None

        with mock.patch("osc_lib.utils.find_resource", mocker):
            columns, data = self.cmd.take_action(parsed_args)

        self.projects_mock.list.assert_called_with(
            domain=self.project.domain_id
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, tuple(data))

    def test_project_list_parent(self):
        self.parent = identity_fakes.FakeProject.create_one_project()
        self.project = identity_fakes.FakeProject.create_one_project(
            attrs={'domain_id': self.domain.id, 'parent_id': self.parent.id}
        )

        arglist = [
            '--parent',
            self.parent.id,
        ]
        verifylist = [
            ('parent', self.parent.id),
        ]

        self.projects_mock.get.return_value = self.parent

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        self.projects_mock.list.assert_called_with(parent=self.parent.id)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, tuple(data))

    def test_project_list_sort(self):
        self.projects_mock.list.return_value = self.projects

        arglist = [
            '--sort',
            'name:asc',
        ]
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        (columns, data) = self.cmd.take_action(parsed_args)
        self.projects_mock.list.assert_called_with()

        collist = ('ID', 'Name')
        self.assertEqual(collist, columns)

        if self.projects[0].name > self.projects[1].name:
            datalists = (
                (self.projects[1].id, self.projects[1].name),
                (self.projects[0].id, self.projects[0].name),
            )
        else:
            datalists = (
                (self.projects[0].id, self.projects[0].name),
                (self.projects[1].id, self.projects[1].name),
            )

        self.assertEqual(datalists, tuple(data))

    def test_project_list_my_projects(self):
        auth_ref = identity_fakes.fake_auth_ref(
            identity_fakes.TOKEN_WITH_PROJECT_ID,
        )
        self.app.client_manager.auth_ref = auth_ref

        arglist = [
            '--my-projects',
        ]
        verifylist = [
            ('my_projects', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)
        self.projects_mock.list.assert_called_with(
            user=self.app.client_manager.auth_ref.user_id
        )

        collist = ('ID', 'Name')
        self.assertEqual(collist, columns)
        datalist = (
            (
                self.project.id,
                self.project.name,
            ),
        )
        self.assertEqual(datalist, tuple(data))

    def test_project_list_with_option_enabled(self):
        arglist = ['--enabled']
        verifylist = [('is_enabled', True)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        kwargs = {'is_enabled': True}
        self.projects_mock.list.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, tuple(data))


class TestProjectSet(TestProject):
    domain = identity_fakes.FakeDomain.create_one_domain()
    project = identity_fakes.FakeProject.create_one_project(
        attrs={'domain_id': domain.id, 'tags': ['tag1', 'tag2', 'tag3']}
    )

    def setUp(self):
        super().setUp()

        self.domains_mock.get.return_value = self.domain

        self.projects_mock.get.return_value = self.project
        self.projects_mock.update.return_value = self.project

        # Get the command object to test
        self.cmd = project.SetProject(self.app, None)

    def test_project_set_no_options(self):
        arglist = [
            self.project.name,
        ]
        verifylist = [
            ('project', self.project.name),
            ('enabled', None),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.assertIsNone(result)

    def test_project_set_name(self):
        arglist = [
            '--name',
            'qwerty',
            '--domain',
            self.project.domain_id,
            self.project.name,
        ]
        verifylist = [
            ('name', 'qwerty'),
            ('domain', self.project.domain_id),
            ('enabled', None),
            ('project', self.project.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': 'qwerty',
        }
        # ProjectManager.update(project, name=, domain=, description=,
        #                       enabled=, **kwargs)
        self.projects_mock.update.assert_called_with(self.project.id, **kwargs)
        self.assertIsNone(result)

    def test_project_set_description(self):
        arglist = [
            '--domain',
            self.project.domain_id,
            '--description',
            'new desc',
            self.project.name,
        ]
        verifylist = [
            ('domain', self.project.domain_id),
            ('description', 'new desc'),
            ('enabled', None),
            ('project', self.project.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'description': 'new desc',
        }
        self.projects_mock.update.assert_called_with(self.project.id, **kwargs)
        self.assertIsNone(result)

    def test_project_set_enable(self):
        arglist = [
            '--domain',
            self.project.domain_id,
            '--enable',
            self.project.name,
        ]
        verifylist = [
            ('domain', self.project.domain_id),
            ('enabled', True),
            ('project', self.project.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'enabled': True,
        }
        self.projects_mock.update.assert_called_with(self.project.id, **kwargs)
        self.assertIsNone(result)

    def test_project_set_disable(self):
        arglist = [
            '--domain',
            self.project.domain_id,
            '--disable',
            self.project.name,
        ]
        verifylist = [
            ('domain', self.project.domain_id),
            ('enabled', False),
            ('project', self.project.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'enabled': False,
        }
        self.projects_mock.update.assert_called_with(self.project.id, **kwargs)
        self.assertIsNone(result)

    def test_project_set_property(self):
        arglist = [
            '--domain',
            self.project.domain_id,
            '--property',
            'fee=fi',
            '--property',
            'fo=fum',
            self.project.name,
        ]
        verifylist = [
            ('domain', self.project.domain_id),
            ('properties', {'fee': 'fi', 'fo': 'fum'}),
            ('project', self.project.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'fee': 'fi',
            'fo': 'fum',
        }
        self.projects_mock.update.assert_called_with(self.project.id, **kwargs)
        self.assertIsNone(result)

    def test_project_set_tags(self):
        arglist = [
            '--name',
            'qwerty',
            '--domain',
            self.project.domain_id,
            '--tag',
            'foo',
            self.project.name,
        ]
        verifylist = [
            ('name', 'qwerty'),
            ('domain', self.project.domain_id),
            ('enabled', None),
            ('project', self.project.name),
            ('tags', ['foo']),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values. new tag is added to original tags for update.
        kwargs = {
            'name': 'qwerty',
            'tags': sorted({'tag1', 'tag2', 'tag3', 'foo'}),
        }
        # ProjectManager.update(project, name=, domain=, description=,
        #                       enabled=, **kwargs)
        self.projects_mock.update.assert_called_with(self.project.id, **kwargs)
        self.assertIsNone(result)

    def test_project_remove_tags(self):
        arglist = [
            '--remove-tag',
            'tag1',
            '--remove-tag',
            'tag2',
            self.project.name,
        ]
        verifylist = [
            ('enabled', None),
            ('project', self.project.name),
            ('remove_tags', ['tag1', 'tag2']),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        kwargs = {'tags': list({'tag3'})}
        self.projects_mock.update.assert_called_with(self.project.id, **kwargs)
        self.assertIsNone(result)

    def test_project_set_with_immutable_option(self):
        arglist = [
            '--domain',
            self.project.domain_id,
            '--immutable',
            self.project.name,
        ]
        verifylist = [
            ('domain', self.project.domain_id),
            ('immutable', True),
            ('enabled', None),
            ('project', self.project.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'options': {'immutable': True},
        }
        self.projects_mock.update.assert_called_with(self.project.id, **kwargs)
        self.assertIsNone(result)

    def test_project_set_with_no_immutable_option(self):
        arglist = [
            '--domain',
            self.project.domain_id,
            '--no-immutable',
            self.project.name,
        ]
        verifylist = [
            ('domain', self.project.domain_id),
            ('immutable', False),
            ('enabled', None),
            ('project', self.project.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'options': {'immutable': False},
        }
        self.projects_mock.update.assert_called_with(self.project.id, **kwargs)
        self.assertIsNone(result)


class TestProjectShow(TestProject):
    domain = identity_fakes.FakeDomain.create_one_domain()

    def setUp(self):
        super().setUp()

        self.project = identity_fakes.FakeProject.create_one_project(
            attrs={'domain_id': self.domain.id}
        )

        # Get the command object to test
        self.cmd = project.ShowProject(self.app, None)

    def test_project_show(self):
        self.projects_mock.get.return_value = self.project

        arglist = [
            self.project.id,
        ]
        verifylist = [
            ('project', self.project.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.identity_client.tokens.get_token_data.return_value = {
            'token': {
                'project': {
                    'domain': {},
                    'name': parsed_args.project,
                    'id': parsed_args.project,
                }
            }
        }

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        self.projects_mock.get.assert_called_once_with(self.project.id)

        collist = (
            'description',
            'domain_id',
            'enabled',
            'id',
            'is_domain',
            'name',
            'parent_id',
            'tags',
        )
        self.assertEqual(collist, columns)
        datalist = (
            self.project.description,
            self.project.domain_id,
            True,
            self.project.id,
            False,
            self.project.name,
            self.project.parent_id,
            self.project.tags,
        )
        self.assertEqual(datalist, data)

    def test_project_show_parents(self):
        self.project = identity_fakes.FakeProject.create_one_project(
            attrs={
                'parent_id': self.project.parent_id,
                'parents': [{'project': {'id': self.project.parent_id}}],
            }
        )
        self.projects_mock.get.return_value = self.project

        arglist = [
            self.project.id,
            '--parents',
        ]
        verifylist = [
            ('project', self.project.id),
            ('parents', True),
            ('children', False),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.identity_client.tokens.get_token_data.return_value = {
            'token': {
                'project': {
                    'domain': {},
                    'name': parsed_args.project,
                    'id': parsed_args.project,
                }
            }
        }

        columns, data = self.cmd.take_action(parsed_args)

        self.projects_mock.get.assert_has_calls(
            [
                call(self.project.id),
                call(
                    self.project.id,
                    parents_as_ids=True,
                    subtree_as_ids=False,
                ),
            ]
        )

        collist = (
            'description',
            'domain_id',
            'enabled',
            'id',
            'is_domain',
            'name',
            'parent_id',
            'parents',
            'tags',
        )
        self.assertEqual(columns, collist)
        datalist = (
            self.project.description,
            self.project.domain_id,
            self.project.enabled,
            self.project.id,
            self.project.is_domain,
            self.project.name,
            self.project.parent_id,
            [{'project': {'id': self.project.parent_id}}],
            self.project.tags,
        )
        self.assertEqual(data, datalist)

    def test_project_show_subtree(self):
        self.project = identity_fakes.FakeProject.create_one_project(
            attrs={
                'parent_id': self.project.parent_id,
                'subtree': [{'project': {'id': 'children-id'}}],
            }
        )
        self.projects_mock.get.return_value = self.project

        arglist = [
            self.project.id,
            '--children',
        ]
        verifylist = [
            ('project', self.project.id),
            ('parents', False),
            ('children', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.identity_client.tokens.get_token_data.return_value = {
            'token': {
                'project': {
                    'domain': {},
                    'name': parsed_args.project,
                    'id': parsed_args.project,
                }
            }
        }

        columns, data = self.cmd.take_action(parsed_args)
        self.projects_mock.get.assert_has_calls(
            [
                call(self.project.id),
                call(
                    self.project.id,
                    parents_as_ids=False,
                    subtree_as_ids=True,
                ),
            ]
        )

        collist = (
            'description',
            'domain_id',
            'enabled',
            'id',
            'is_domain',
            'name',
            'parent_id',
            'subtree',
            'tags',
        )
        self.assertEqual(columns, collist)
        datalist = (
            self.project.description,
            self.project.domain_id,
            self.project.enabled,
            self.project.id,
            self.project.is_domain,
            self.project.name,
            self.project.parent_id,
            [{'project': {'id': 'children-id'}}],
            self.project.tags,
        )
        self.assertEqual(data, datalist)

    def test_project_show_parents_and_children(self):
        self.project = identity_fakes.FakeProject.create_one_project(
            attrs={
                'parent_id': self.project.parent_id,
                'parents': [{'project': {'id': self.project.parent_id}}],
                'subtree': [{'project': {'id': 'children-id'}}],
            }
        )
        self.projects_mock.get.return_value = self.project

        arglist = [
            self.project.id,
            '--parents',
            '--children',
        ]
        verifylist = [
            ('project', self.project.id),
            ('parents', True),
            ('children', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.identity_client.tokens.get_token_data.return_value = {
            'token': {
                'project': {
                    'domain': {},
                    'name': parsed_args.project,
                    'id': parsed_args.project,
                }
            }
        }

        columns, data = self.cmd.take_action(parsed_args)
        self.projects_mock.get.assert_has_calls(
            [
                call(self.project.id),
                call(
                    self.project.id,
                    parents_as_ids=True,
                    subtree_as_ids=True,
                ),
            ]
        )

        collist = (
            'description',
            'domain_id',
            'enabled',
            'id',
            'is_domain',
            'name',
            'parent_id',
            'parents',
            'subtree',
            'tags',
        )
        self.assertEqual(columns, collist)
        datalist = (
            self.project.description,
            self.project.domain_id,
            self.project.enabled,
            self.project.id,
            self.project.is_domain,
            self.project.name,
            self.project.parent_id,
            [{'project': {'id': self.project.parent_id}}],
            [{'project': {'id': 'children-id'}}],
            self.project.tags,
        )
        self.assertEqual(data, datalist)

    def test_project_show_with_domain(self):
        project = identity_fakes.FakeProject.create_one_project(
            {"name": self.project.name}
        )

        self.identity_client.tokens.get_token_data.return_value = {
            'token': {
                'project': {
                    'domain': {"id": self.project.domain_id},
                    'name': self.project.name,
                    'id': self.project.id,
                }
            }
        }

        identity_client = self.identity_client
        arglist = [
            "--domain",
            self.domain.id,
            project.name,
        ]
        verifylist = [
            ('domain', self.domain.id),
            ('project', project.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        project_str = common._get_token_resource(
            identity_client, 'project', parsed_args.project, parsed_args.domain
        )
        self.assertEqual(self.project.id, project_str)

        arglist = [
            "--domain",
            project.domain_id,
            project.name,
        ]
        verifylist = [
            ('domain', project.domain_id),
            ('project', project.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        project_str = common._get_token_resource(
            identity_client, 'project', parsed_args.project, parsed_args.domain
        )
        self.assertEqual(project.name, project_str)



# Copyright 2016 ZTE Corporation.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""Network trunk and subports action implementations"""

import logging
import typing as ty

from cliff import columns as cliff_columns
from osc_lib.cli import format_columns
from osc_lib.cli import identity as identity_utils
from osc_lib.cli import parseractions
from osc_lib import exceptions
from osc_lib import utils as osc_utils

from openstackclient import command
from openstackclient.i18n import _

LOG = logging.getLogger(__name__)

TRUNK = 'trunk'
TRUNKS = 'trunks'
SUB_PORTS = 'sub_ports'


class AdminStateColumn(cliff_columns.FormattableColumn[bool]):
    def human_readable(self):
        return 'UP' if self._value else 'DOWN'


class CreateNetworkTrunk(command.ShowOne):
    """Create a network trunk for a given project"""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'name', metavar='<name>', help=_("Name of the trunk to create")
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_("A description of the trunk"),
        )
        parser.add_argument(
            '--parent-port',
            metavar='<parent-port>',
            required=True,
            help=_("Parent port belonging to this trunk (name or ID)"),
        )
        parser.add_argument(
            '--subport',
            metavar='<port=,segmentation-type=,segmentation-id=>',
            action=parseractions.MultiKeyValueAction,
            dest='add_subports',
            optional_keys=['segmentation-id', 'segmentation-type'],
            required_keys=['port'],
            help=_(
                "Subport to add. Subport is of form "
                "'port=<name or ID>,segmentation-type=<segmentation-type>,"
                "segmentation-id=<segmentation-ID>' (repeat option "
                "to add multiple subports)"
            ),
        )
        admin_group = parser.add_mutually_exclusive_group()
        admin_group.add_argument(
            '--enable',
            action='store_true',
            default=True,
            help=_("Enable trunk (default)"),
        )
        admin_group.add_argument(
            '--disable', action='store_true', help=_("Disable trunk")
        )
        identity_utils.add_project_owner_option_to_parser(parser)
        return parser

    def take_action(self, parsed_args):
        network_client = self.app.client_manager.network
        identity_client = self.app.client_manager.identity
        attrs = _get_attrs_for_trunk(
            network_client, identity_client, parsed_args
        )
        obj = network_client.create_trunk(**attrs)
        display_columns, columns = _get_columns(obj)
        data = osc_utils.get_dict_properties(
            obj, columns, formatters=_formatters
        )
        return display_columns, data


class DeleteNetworkTrunk(command.Command):
    """Delete a given network trunk"""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'trunk',
            metavar="<trunk>",
            nargs="+",
            help=_("Trunk(s) to delete (name or ID)"),
        )
        return parser

    def take_action(self, parsed_args):
        network_client = self.app.client_manager.network
        result = 0
        for trunk in parsed_args.trunk:
            try:
                trunk_id = network_client.find_trunk(
                    trunk,
                    ignore_missing=False,
                ).id
                network_client.delete_trunk(trunk_id)
            except Exception as e:
                result += 1
                LOG.error(
                    _(
                        "Failed to delete trunk with name "
                        "or ID '%(trunk)s': %(e)s"
                    ),
                    {'trunk': trunk, 'e': e},
                )
        if result > 0:
            total = len(parsed_args.trunk)
            msg = _("%(result)s of %(total)s trunks failed to delete.") % {
                'result': result,
                'total': total,
            }
            raise exceptions.CommandError(msg)


class ListNetworkTrunk(command.Lister):
    """List all network trunks"""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--long',
            action='store_true',
            default=False,
            help=_("List additional fields in output"),
        )
        return parser

    def take_action(self, parsed_args):
        network_client = self.app.client_manager.network
        data = network_client.trunks()
        headers: tuple[str, ...] = ('ID', 'Name', 'Parent Port', 'Description')
        columns: tuple[str, ...] = ('id', 'name', 'port_id', 'description')
        if parsed_args.long:
            headers += (
                'Status',
                'State',
                'Created At',
                'Updated At',
            )
            columns += ('status', 'admin_state_up', 'created_at', 'updated_at')
        return (
            headers,
            (
                osc_utils.get_item_properties(
                    s,
                    columns,
                    formatters=_formatters,
                )
                for s in data
            ),
        )


class SetNetworkTrunk(command.Command):
    """Set network trunk properties"""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'trunk', metavar="<trunk>", help=_("Trunk to modify (name or ID)")
        )
        parser.add_argument(
            '--name', metavar="<name>", help=_("Set trunk name")
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_("A description of the trunk"),
        )
        parser.add_argument(
            '--subport',
            metavar='<port=,segmentation-type=,segmentation-id=>',
            action=parseractions.MultiKeyValueAction,
            dest='set_subports',
            optional_keys=['segmentation-id', 'segmentation-type'],
            required_keys=['port'],
            help=_(
                "Subport to add. Subport is of form "
                "'port=<name or ID>,segmentation-type=<segmentation-type>,"
                "segmentation-id=<segmentation-ID>' (repeat option "
                "to add multiple subports)"
            ),
        )
        admin_group = parser.add_mutually_exclusive_group()
        admin_group.add_argument(
            '--enable', action='store_true', help=_("Enable trunk")
        )
        admin_group.add_argument(
            '--disable', action='store_true', help=_("Disable trunk")
        )
        return parser

    def take_action(self, parsed_args):
        network_client = self.app.client_manager.network
        identity_client = self.app.client_manager.identity
        trunk_id = network_client.find_trunk(
            parsed_args.trunk,
            ignore_missing=False,
        )
        attrs = _get_attrs_for_trunk(
            network_client, identity_client, parsed_args
        )
        try:
            network_client.update_trunk(trunk_id, **attrs)
        except Exception as e:
            msg = _("Failed to set trunk '%(t)s': %(e)s") % {
                't': parsed_args.trunk,
                'e': e,
            }
            raise exceptions.CommandError(msg)
        if parsed_args.set_subports:
            subport_attrs = _get_attrs_for_subports(
                network_client, parsed_args
            )
            try:
                network_client.add_trunk_subports(trunk_id, subport_attrs)
            except Exception as e:
                msg = _("Failed to add subports to trunk '%(t)s': %(e)s") % {
                    't': parsed_args.trunk,
                    'e': e,
                }
                raise exceptions.CommandError(msg)


class ShowNetworkTrunk(command.ShowOne):
    """Show information of a given network trunk"""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'trunk', metavar="<trunk>", help=_("Trunk to display (name or ID)")
        )
        return parser

    def take_action(self, parsed_args):
        network_client = self.app.client_manager.network
        trunk_id = network_client.find_trunk(
            parsed_args.trunk,
            ignore_missing=False,
        ).id
        obj = network_client.get_trunk(trunk_id)
        display_columns, columns = _get_columns(obj)
        data = osc_utils.get_dict_properties(
            obj, columns, formatters=_formatters
        )
        return display_columns, data


class ListNetworkSubport(command.Lister):
    """List all subports for a given network trunk"""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--trunk',
            required=True,
            metavar="<trunk>",
            help=_("List only subports belonging to this trunk (name or ID)"),
        )
        return parser

    def take_action(self, parsed_args):
        network_client = self.app.client_manager.network
        trunk_id = network_client.find_trunk(
            parsed_args.trunk,
            ignore_missing=False,
        )
        data = network_client.get_trunk_subports(trunk_id)
        headers: tuple[str, ...] = (
            'Port',
            'Segmentation Type',
            'Segmentation ID',
        )
        columns: tuple[str, ...] = (
            'port_id',
            'segmentation_type',
            'segmentation_id',
        )
        return (
            headers,
            (
                osc_utils.get_dict_properties(
                    s,
                    columns,
                )
                for s in data[SUB_PORTS]
            ),
        )


class UnsetNetworkTrunk(command.Command):
    """Unset subports from a given network trunk"""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'trunk',
            metavar="<trunk>",
            help=_("Unset subports from this trunk (name or ID)"),
        )
        parser.add_argument(
            '--subport',
            metavar="<subport>",
            required=True,
            action='append',
            dest='unset_subports',
            help=_(
                "Subport to unset (name or ID of the port) "
                "(repeat option to unset multiple subports)"
            ),
        )
        return parser

    def take_action(self, parsed_args):
        network_client = self.app.client_manager.network
        attrs = _get_attrs_for_subports(network_client, parsed_args)
        trunk_id = network_client.find_trunk(
            parsed_args.trunk,
            ignore_missing=False,
        )
        network_client.delete_trunk_subports(trunk_id, attrs)


_formatters = {
    'admin_state_up': AdminStateColumn,
    'sub_ports': format_columns.ListDictColumn,
}


def _get_columns(item):
    hidden_columns = ['location', 'tenant_id']
    return osc_utils.get_osc_show_columns_for_sdk_resource(
        item, {}, hidden_columns
    )


def _get_attrs_for_trunk(network_client, identity_client, parsed_args):
    attrs: dict[str, ty.Any] = {}
    if parsed_args.name is not None:
        attrs['name'] = str(parsed_args.name)
    if parsed_args.description is not None:
        attrs['description'] = str(parsed_args.description)
    if parsed_args.enable:
        attrs['admin_state_up'] = True
    if parsed_args.disable:
        attrs['admin_state_up'] = False
    if 'parent_port' in parsed_args and parsed_args.parent_port is not None:
        port_id = network_client.find_port(
            parsed_args.parent_port,
            ignore_missing=False,
        ).id
        attrs['port_id'] = port_id
    if 'add_subports' in parsed_args and parsed_args.add_subports is not None:
        attrs[SUB_PORTS] = _format_subports(
            network_client, parsed_args.add_subports
        )

    # "trunk set" command doesn't support setting project.
    if 'project' in parsed_args and parsed_args.project is not None:
        project_id = identity_utils.find_project(
            identity_client,
            parsed_args.project,
            parsed_args.project_domain,
        ).id
        attrs['tenant_id'] = project_id

    return attrs


def _format_subports(network_client, subports):
    attrs = []
    for subport in subports:
        subport_attrs = {}
        if subport.get('port'):
            port_id = network_client.find_port(
                subport['port'],
                ignore_missing=False,
            ).id
            subport_attrs['port_id'] = port_id
        if subport.get('segmentation-id'):
            try:
                subport_attrs['segmentation_id'] = int(
                    subport['segmentation-id']
                )
            except ValueError:
                msg = (
                    _("Segmentation-id '%s' is not an integer")
                    % subport['segmentation-id']
                )
                raise exceptions.CommandError(msg)
        if subport.get('segmentation-type'):
            subport_attrs['segmentation_type'] = subport['segmentation-type']
        attrs.append(subport_attrs)
    return attrs


def _get_attrs_for_subports(network_client, parsed_args):
    attrs = []
    if 'set_subports' in parsed_args and parsed_args.set_subports is not None:
        attrs = _format_subports(network_client, parsed_args.set_subports)
    if (
        'unset_subports' in parsed_args
        and parsed_args.unset_subports is not None
    ):
        subports_list = []
        for subport in parsed_args.unset_subports:
            port_id = network_client.find_port(
                subport,
                ignore_missing=False,
            )['id']
            subports_list.append({'port_id': port_id})
        attrs = subports_list
    return attrs



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import itertools
import json
import time
import uuid

from tempest.lib import exceptions

from openstackclient.compute.v2 import server as v2_server
from openstackclient.tests.functional.compute.v2 import common
from openstackclient.tests.functional.volume.v2 import common as volume_common


class ServerTests(common.ComputeTestCase):
    """Functional tests for openstack server commands"""

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.haz_network = cls.is_service_enabled('network')

    def test_server_list(self):
        """Test server list"""
        cmd_output = self.server_create()
        name1 = cmd_output['name']
        cmd_output = self.server_create()
        name2 = cmd_output['name']
        self.wait_for_status(name1, "ACTIVE")
        self.wait_for_status(name2, "ACTIVE")

        cmd_output = self.openstack(
            'server list',
            parse_output=True,
        )
        col_name = [x["Name"] for x in cmd_output]
        self.assertIn(name1, col_name)
        self.assertIn(name2, col_name)

        # Test list --status PAUSED
        raw_output = self.openstack('server pause ' + name2)
        self.assertEqual("", raw_output)
        self.wait_for_status(name2, "PAUSED")
        cmd_output = self.openstack(
            'server list ' + '--status ACTIVE',
            parse_output=True,
        )
        col_name = [x["Name"] for x in cmd_output]
        self.assertIn(name1, col_name)
        self.assertNotIn(name2, col_name)
        cmd_output = self.openstack(
            'server list ' + '--status PAUSED',
            parse_output=True,
        )
        col_name = [x["Name"] for x in cmd_output]
        self.assertNotIn(name1, col_name)
        self.assertIn(name2, col_name)

    def test_server_list_with_marker_and_deleted(self):
        """Test server list with deleted and marker"""
        cmd_output = self.server_create(cleanup=False)
        name1 = cmd_output['name']
        cmd_output = self.server_create(cleanup=False)
        name2 = cmd_output['name']
        id2 = cmd_output['id']
        self.wait_for_status(name1, "ACTIVE")
        self.wait_for_status(name2, "ACTIVE")

        # Test list --marker with ID
        cmd_output = self.openstack(
            'server list --marker ' + id2,
            parse_output=True,
        )
        col_name = [x["Name"] for x in cmd_output]
        self.assertIn(name1, col_name)

        # Test list --marker with Name
        cmd_output = self.openstack(
            'server list --marker ' + name2,
            parse_output=True,
        )
        col_name = [x["Name"] for x in cmd_output]
        self.assertIn(name1, col_name)

        self.openstack('server delete --wait ' + name1)
        self.openstack('server delete --wait ' + name2)

        # Test list --deleted --marker with ID
        cmd_output = self.openstack(
            'server list --deleted --marker ' + id2,
            parse_output=True,
        )
        col_name = [x["Name"] for x in cmd_output]
        self.assertIn(name1, col_name)

        # Test list --deleted --marker with Name
        try:
            cmd_output = self.openstack(
                'server list --deleted --marker ' + name2,
                parse_output=True,
            )
        except exceptions.CommandFailed as e:
            self.assertIn(
                f'marker [{name2}] not found', e.stderr.decode('utf-8')
            )

    def test_server_list_with_changes_before(self):
        """Test server list.

        Getting the servers list with updated_at time equal or
        before than changes-before.
        """
        cmd_output = self.server_create()
        server_name1 = cmd_output['name']

        cmd_output = self.server_create()
        server_name2 = cmd_output['name']
        updated_at2 = cmd_output['updated']

        cmd_output = self.server_create()
        server_name3 = cmd_output['name']

        cmd_output = self.openstack(
            '--os-compute-api-version 2.66 ' + 'server list '
            '--changes-before ' + updated_at2,
            parse_output=True,
        )

        col_updated = [server["Name"] for server in cmd_output]
        self.assertIn(server_name1, col_updated)
        self.assertIn(server_name2, col_updated)
        self.assertNotIn(server_name3, col_updated)

    def test_server_list_with_changes_since(self):
        """Test server list.

        Getting the servers list with updated_at time equal or
        later than changes-since.
        """
        cmd_output = self.server_create()
        server_name1 = cmd_output['name']
        cmd_output = self.server_create()
        server_name2 = cmd_output['name']
        updated_at2 = cmd_output['updated']
        cmd_output = self.server_create()
        server_name3 = cmd_output['name']

        cmd_output = self.openstack(
            'server list --changes-since ' + updated_at2,
            parse_output=True,
        )

        col_updated = [server["Name"] for server in cmd_output]
        self.assertNotIn(server_name1, col_updated)
        self.assertIn(server_name2, col_updated)
        self.assertIn(server_name3, col_updated)

    def test_server_list_with_changes_before_and_changes_since(self):
        """Test server list.

        Getting the servers list with updated_at time equal or before than
        changes-before and equal or later than changes-since.
        """
        cmd_output = self.server_create()
        server_name1 = cmd_output['name']
        cmd_output = self.server_create()
        server_name2 = cmd_output['name']
        updated_at2 = cmd_output['updated']
        cmd_output = self.server_create()
        server_name3 = cmd_output['name']
        updated_at3 = cmd_output['updated']

        cmd_output = self.openstack(
            '--os-compute-api-version 2.66 '
            + 'server list '
            + '--changes-since '
            + updated_at2
            + ' --changes-before '
            + updated_at3,
            parse_output=True,
        )

        col_updated = [server["Name"] for server in cmd_output]
        self.assertNotIn(server_name1, col_updated)
        self.assertIn(server_name2, col_updated)
        self.assertIn(server_name3, col_updated)

    def test_server_set(self):
        """Test server create, delete, set, show"""
        cmd_output = self.server_create()
        name = cmd_output['name']
        # self.wait_for_status(name, "ACTIVE")

        # Have a look at some other fields
        flavor = self.openstack(
            'flavor show ' + self.flavor_name,
            parse_output=True,
        )
        self.assertEqual(
            self.flavor_name,
            flavor['name'],
        )
        # assume the v2.47+ output format
        self.assertIsInstance(cmd_output['flavor'], dict)
        self.assertIn('name', cmd_output['flavor'])
        self.assertEqual(flavor['name'], cmd_output['flavor']['name'])
        image = self.openstack(
            'image show ' + self.image_name,
            parse_output=True,
        )
        self.assertEqual(
            self.image_name,
            image['name'],
        )
        self.assertEqual(
            '{} ({})'.format(image['name'], image['id']),
            cmd_output["image"],
        )

        # Test properties set
        raw_output = self.openstack(
            'server set ' + '--property a=b --property c=d ' + name
        )
        self.assertOutput('', raw_output)

        cmd_output = self.openstack(
            'server show ' + name,
            parse_output=True,
        )
        # Really, shouldn't this be a list?
        self.assertEqual(
            {'a': 'b', 'c': 'd'},
            cmd_output['properties'],
        )

        raw_output = self.openstack('server unset ' + '--property a ' + name)
        cmd_output = self.openstack(
            'server show ' + name,
            parse_output=True,
        )
        self.assertEqual(
            {'c': 'd'},
            cmd_output['properties'],
        )

        # Test set --name
        new_name = uuid.uuid4().hex
        raw_output = self.openstack(
            'server set ' + '--name ' + new_name + ' ' + name
        )
        self.assertOutput("", raw_output)
        cmd_output = self.openstack(
            'server show ' + new_name,
            parse_output=True,
        )
        self.assertEqual(
            new_name,
            cmd_output["name"],
        )
        # Put it back so we clean up properly
        raw_output = self.openstack(
            'server set ' + '--name ' + name + ' ' + new_name
        )
        self.assertOutput("", raw_output)

    def test_server_show(self):
        """Test server show"""
        cmd_output = self.server_create()
        name = cmd_output['name']

        # Simple show
        cmd_output = json.loads(self.openstack(f'server show -f json {name}'))
        self.assertEqual(
            name,
            cmd_output["name"],
        )

        # Show diagnostics
        cmd_output = json.loads(
            self.openstack(f'server show -f json {name} --diagnostics')
        )
        self.assertIn('driver', cmd_output)

        # Show topology
        cmd_output = json.loads(
            self.openstack(
                f'server show -f json {name} --topology '
                f'--os-compute-api-version 2.78'
            )
        )
        self.assertIn('topology', cmd_output)

    def test_server_actions(self):
        """Test server action pairs

        suspend/resume
        pause/unpause
        rescue/unrescue
        lock/unlock
        """
        cmd_output = self.server_create()
        name = cmd_output['name']

        # suspend
        raw_output = self.openstack('server suspend ' + name)
        self.assertEqual("", raw_output)
        self.wait_for_status(name, "SUSPENDED")

        # resume
        raw_output = self.openstack('server resume ' + name)
        self.assertEqual("", raw_output)
        self.wait_for_status(name, "ACTIVE")

        # pause
        raw_output = self.openstack('server pause ' + name)
        self.assertEqual("", raw_output)
        self.wait_for_status(name, "PAUSED")

        # unpause
        raw_output = self.openstack('server unpause ' + name)
        self.assertEqual("", raw_output)
        self.wait_for_status(name, "ACTIVE")

        # rescue
        raw_output = self.openstack('server rescue ' + name)
        self.assertEqual("", raw_output)
        self.wait_for_status(name, "RESCUE")

        # unrescue
        raw_output = self.openstack('server unrescue ' + name)
        self.assertEqual("", raw_output)
        self.wait_for_status(name, "ACTIVE")

        # rescue with image
        raw_output = self.openstack(
            'server rescue --image ' + self.image_name + ' ' + name
        )
        self.assertEqual("", raw_output)
        self.wait_for_status(name, "RESCUE")

        # unrescue
        raw_output = self.openstack('server unrescue ' + name)
        self.assertEqual("", raw_output)
        self.wait_for_status(name, "ACTIVE")

        # lock
        raw_output = self.openstack('server lock ' + name)
        self.assertEqual("", raw_output)
        # NOTE(dtroyer): No way to verify this status???

        # unlock
        raw_output = self.openstack('server unlock ' + name)
        self.assertEqual("", raw_output)
        # NOTE(dtroyer): No way to verify this status???

    def test_server_attach_detach_floating_ip(self):
        """Test floating ip create/delete; server add/remove floating ip"""
        if not self.haz_network:
            # NOTE(dtroyer): As of Ocata release Nova forces nova-network to
            #                run in a cells v1 configuration.  Floating IP
            #                and network functions currently do not work in
            #                the gate jobs so we have to skip this.  It is
            #                known to work tested against a Mitaka nova-net
            #                DevStack without cells.
            self.skipTest("No Network service present")

        def _chain_addresses(addresses):
            # Flatten a dict of lists mapping network names to IP addresses,
            # returning only the IP addresses
            #
            # >>> _chain_addresses({'private': ['10.1.0.32', '172.24.5.41']})
            # ['10.1.0.32', '172.24.5.41']
            return itertools.chain(*[*addresses.values()])

        cmd_output = self.server_create()
        name = cmd_output['name']
        self.wait_for_status(name, "ACTIVE")

        # attach ip
        cmd_output = self.openstack(
            'floating ip create ' + 'public',
            parse_output=True,
        )

        # Look for Neutron value first, then nova-net
        floating_ip = cmd_output.get(
            'floating_ip_address',
            cmd_output.get(
                'ip',
                None,
            ),
        )
        self.assertNotEqual('', cmd_output['id'])
        self.assertNotEqual('', floating_ip)
        self.addCleanup(
            self.openstack, 'floating ip delete ' + str(cmd_output['id'])
        )

        raw_output = self.openstack(
            'server add floating ip ' + name + ' ' + floating_ip
        )
        self.assertEqual("", raw_output)

        # Loop a few times since this is timing-sensitive
        # Just hard-code it for now, since there is no pause and it is
        # racy we shouldn't have to wait too long, a minute seems reasonable
        wait_time = 0
        while wait_time < 60:
            cmd_output = self.openstack(
                'server show ' + name,
                parse_output=True,
            )
            if floating_ip not in _chain_addresses(cmd_output['addresses']):
                # Hang out for a bit and try again
                print('retrying floating IP check')
                wait_time += 10
                time.sleep(10)
            else:
                break

        self.assertIn(
            floating_ip,
            _chain_addresses(cmd_output['addresses']),
        )

        # detach ip
        raw_output = self.openstack(
            'server remove floating ip ' + name + ' ' + floating_ip
        )
        self.assertEqual("", raw_output)

        # Loop a few times since this is timing-sensitive
        # Just hard-code it for now, since there is no pause and it is
        # racy we shouldn't have to wait too long, a minute seems reasonable
        wait_time = 0
        while wait_time < 60:
            cmd_output = self.openstack(
                'server show ' + name,
                parse_output=True,
            )
            if floating_ip in _chain_addresses(cmd_output['addresses']):
                # Hang out for a bit and try again
                print('retrying floating IP check')
                wait_time += 10
                time.sleep(10)
            else:
                break

        cmd_output = self.openstack(
            'server show ' + name,
            parse_output=True,
        )
        self.assertNotIn(
            floating_ip,
            _chain_addresses(cmd_output['addresses']),
        )

    def test_server_reboot(self):
        """Test server reboot"""
        cmd_output = self.server_create()
        name = cmd_output['name']

        # reboot
        raw_output = self.openstack('server reboot ' + name)
        self.assertEqual("", raw_output)
        self.wait_for_status(name, "ACTIVE")

    def test_server_boot_from_volume(self):
        """Test server create from volume, server delete"""
        # get volume status wait function
        volume_wait_for = volume_common.BaseVolumeTests.wait_for_status

        # get image size
        cmd_output = self.openstack(
            'image show ' + self.image_name,
            parse_output=True,
        )
        try:
            image_size = cmd_output['min_disk']
            if image_size < 1:
                image_size = 1
        except ValueError:
            image_size = 1

        # create volume from image
        volume_name = uuid.uuid4().hex
        cmd_output = self.openstack(
            'volume create '
            + '--image '
            + self.image_name
            + ' '
            + '--size '
            + str(image_size)
            + ' '
            + volume_name,
            parse_output=True,
        )
        self.assertIsNotNone(cmd_output["id"])
        self.addCleanup(self.openstack, 'volume delete ' + volume_name)
        self.assertEqual(
            volume_name,
            cmd_output['name'],
        )
        volume_wait_for("volume", volume_name, "available")

        # create empty volume
        empty_volume_name = uuid.uuid4().hex
        cmd_output = self.openstack(
            'volume create '
            + '--size '
            + str(image_size)
            + ' '
            + empty_volume_name,
            parse_output=True,
        )
        self.assertIsNotNone(cmd_output["id"])
        self.addCleanup(self.openstack, 'volume delete ' + empty_volume_name)
        self.assertEqual(
            empty_volume_name,
            cmd_output['name'],
        )
        volume_wait_for("volume", empty_volume_name, "available")

        # create server
        server_name = uuid.uuid4().hex
        server = self.openstack(
            'server create '
            + '--flavor '
            + self.flavor_name
            + ' '
            + '--volume '
            + volume_name
            + ' '
            + '--block-device-mapping vdb='
            + empty_volume_name
            + ' '
            + self.network_arg
            + ' '
            + '--wait '
            + server_name,
            parse_output=True,
        )
        self.assertIsNotNone(server["id"])
        self.addCleanup(self.openstack, 'server delete --wait ' + server_name)
        self.assertEqual(
            server_name,
            server['name'],
        )

        # check that image indicates server "booted from volume"
        self.assertEqual(
            v2_server.IMAGE_STRING_FOR_BFV,
            server['image'],
        )
        # check server list too
        servers = self.openstack(
            'server list',
            parse_output=True,
        )
        self.assertEqual(v2_server.IMAGE_STRING_FOR_BFV, servers[0]['Image'])

        # check volumes
        cmd_output = self.openstack(
            'volume show ' + volume_name,
            parse_output=True,
        )
        attachments = cmd_output['attachments']
        self.assertEqual(
            1,
            len(attachments),
        )
        self.assertEqual(
            server['id'],
            attachments[0]['server_id'],
        )
        self.assertEqual(
            "in-use",
            cmd_output['status'],
        )

        # NOTE(dtroyer): Prior to https://review.opendev.org/#/c/407111
        #                --block-device-mapping was ignored if --volume
        #                present on the command line.  Now we should see the
        #                attachment.
        cmd_output = self.openstack(
            'volume show ' + empty_volume_name,
            parse_output=True,
        )
        attachments = cmd_output['attachments']
        self.assertEqual(
            1,
            len(attachments),
        )
        self.assertEqual(
            server['id'],
            attachments[0]['server_id'],
        )
        self.assertEqual(
            "in-use",
            cmd_output['status'],
        )

    def _test_server_boot_with_bdm_volume(self, use_legacy):
        """Test server create from volume, server delete"""
        # get volume status wait function
        volume_wait_for = volume_common.BaseVolumeTests.wait_for_status

        # create source empty volume
        volume_name = uuid.uuid4().hex
        cmd_output = self.openstack(
            'volume create ' + '--size 1 ' + volume_name,
            parse_output=True,
        )
        volume_id = cmd_output["id"]
        self.assertIsNotNone(volume_id)
        self.addCleanup(self.openstack, 'volume delete ' + volume_name)
        self.assertEqual(volume_name, cmd_output['name'])
        volume_wait_for("volume", volume_name, "available")

        if use_legacy:
            bdm_arg = f'--block-device-mapping vdb={volume_name}'
        else:
            bdm_arg = (
                f'--block-device '
                f'device_name=vdb,source_type=volume,boot_index=1,'
                f'uuid={volume_id}'
            )

        # create server
        server_name = uuid.uuid4().hex
        server = self.openstack(
            'server create '
            + '--flavor '
            + self.flavor_name
            + ' '
            + '--image '
            + self.image_name
            + ' '
            + bdm_arg
            + ' '
            + self.network_arg
            + ' '
            + '--wait '
            + server_name,
            parse_output=True,
        )
        self.assertIsNotNone(server["id"])
        self.addCleanup(self.openstack, 'server delete --wait ' + server_name)
        self.assertEqual(
            server_name,
            server['name'],
        )

        # check server volumes_attached, format is
        # {"volumes_attached": "id='2518bc76-bf0b-476e-ad6b-571973745bb5'",}
        cmd_output = self.openstack(
            'server show ' + server_name,
            parse_output=True,
        )
        volumes_attached = cmd_output['volumes_attached']
        self.assertIsNotNone(volumes_attached)

        # check volumes
        cmd_output = self.openstack(
            'volume show ' + volume_name,
            parse_output=True,
        )
        attachments = cmd_output['attachments']
        self.assertEqual(
            1,
            len(attachments),
        )
        self.assertEqual(
            server['id'],
            attachments[0]['server_id'],
        )
        self.assertEqual(
            "in-use",
            cmd_output['status'],
        )

    def test_server_boot_with_bdm_volume(self):
        """Test server create from image with bdm volume, server delete"""
        self._test_server_boot_with_bdm_volume(use_legacy=False)

    # TODO(stephenfin): Remove when we drop support for the
    # '--block-device-mapping' option
    def test_server_boot_with_bdm_volume_legacy(self):
        """Test server create from image with bdm volume, server delete"""
        self._test_server_boot_with_bdm_volume(use_legacy=True)

    def _test_server_boot_with_bdm_snapshot(self, use_legacy):
        """Test server create from image with bdm snapshot, server delete"""
        # get volume status wait function
        volume_wait_for = volume_common.BaseVolumeTests.wait_for_status
        volume_wait_for_delete = volume_common.BaseVolumeTests.wait_for_delete

        # create source empty volume
        empty_volume_name = uuid.uuid4().hex
        cmd_output = self.openstack(
            'volume create ' + '--size 1 ' + empty_volume_name,
            parse_output=True,
        )
        self.assertIsNotNone(cmd_output["id"])
        self.addCleanup(self.openstack, 'volume delete ' + empty_volume_name)
        self.assertEqual(empty_volume_name, cmd_output['name'])
        volume_wait_for("volume", empty_volume_name, "available")

        # create snapshot of source empty volume
        empty_snapshot_name = uuid.uuid4().hex
        cmd_output = self.openstack(
            'volume snapshot create '
            + '--volume '
            + empty_volume_name
            + ' '
            + empty_snapshot_name,
            parse_output=True,
        )
        empty_snapshot_id = cmd_output["id"]
        self.assertIsNotNone(empty_snapshot_id)
        # Deleting volume snapshot take time, so we need to wait until the
        # snapshot goes. Entries registered by self.addCleanup will be called
        # in the reverse order, so we need to register wait_for_delete first.
        self.addCleanup(
            volume_wait_for_delete, 'volume snapshot', empty_snapshot_name
        )
        self.addCleanup(
            self.openstack, 'volume snapshot delete ' + empty_snapshot_name
        )
        self.assertEqual(
            empty_snapshot_name,
            cmd_output['name'],
        )
        volume_wait_for("volume snapshot", empty_snapshot_name, "available")

        if use_legacy:
            bdm_arg = (
                f'--block-device-mapping '
                f'vdb={empty_snapshot_name}:snapshot:1:true'
            )
        else:
            bdm_arg = (
                f'--block-device '
                f'device_name=vdb,uuid={empty_snapshot_id},'
                f'source_type=snapshot,volume_size=1,'
                f'delete_on_termination=true,boot_index=1'
            )

        # create server with bdm snapshot
        server_name = uuid.uuid4().hex
        server = self.openstack(
            'server create '
            + '--flavor '
            + self.flavor_name
            + ' '
            + '--image '
            + self.image_name
            + ' '
            + bdm_arg
            + ' '
            + self.network_arg
            + ' '
            + '--wait '
            + server_name,
            parse_output=True,
        )
        self.assertIsNotNone(server["id"])
        self.assertEqual(
            server_name,
            server['name'],
        )
        self.wait_for_status(server_name, 'ACTIVE')

        # check server volumes_attached, format is
        # {"volumes_attached": "id='2518bc76-bf0b-476e-ad6b-571973745bb5'",}
        cmd_output = self.openstack(
            'server show ' + server_name,
            parse_output=True,
        )
        volumes_attached = cmd_output['volumes_attached']
        self.assertIsNotNone(volumes_attached)
        attached_volume_id = volumes_attached[0]["id"]

        # check the volume that attached on server
        cmd_output = self.openstack(
            'volume show ' + attached_volume_id,
            parse_output=True,
        )
        attachments = cmd_output['attachments']
        self.assertEqual(
            1,
            len(attachments),
        )
        self.assertEqual(
            server['id'],
            attachments[0]['server_id'],
        )
        self.assertEqual(
            "in-use",
            cmd_output['status'],
        )

        # delete server, then check the attached volume had been deleted,
        # <delete-on-terminate>=true
        self.openstack('server delete --wait ' + server_name)
        cmd_output = self.openstack(
            'volume list',
            parse_output=True,
        )
        target_volume = [
            each_volume
            for each_volume in cmd_output
            if each_volume['ID'] == attached_volume_id
        ]
        if target_volume:
            # check the attached volume is 'deleting' status
            self.assertEqual('deleting', target_volume[0]['Status'])
        else:
            # the attached volume had been deleted
            pass

    def test_server_boot_with_bdm_snapshot(self):
        """Test server create from image with bdm snapshot, server delete"""
        self._test_server_boot_with_bdm_snapshot(use_legacy=False)

    # TODO(stephenfin): Remove when we drop support for the
    # '--block-device-mapping' option
    def test_server_boot_with_bdm_snapshot_legacy(self):
        """Test server create from image with bdm snapshot, server delete"""
        self._test_server_boot_with_bdm_snapshot(use_legacy=True)

    def _test_server_boot_with_bdm_image(self, use_legacy):
        # Tests creating a server where the root disk is backed by the given
        # --image but a --block-device-mapping with type=image is provided so
        # that the compute service creates a volume from that image and
        # attaches it as a non-root volume on the server. The block device is
        # marked as delete_on_termination=True so it will be automatically
        # deleted when the server is deleted.

        if use_legacy:
            # This means create a 1GB volume from the specified image, attach
            # it to the server at /dev/vdb and delete the volume when the
            # server is deleted.
            bdm_arg = (
                f'--block-device-mapping vdb={self.image_name}:image:1:true '
            )
        else:
            # get image ID
            cmd_output = self.openstack(
                'image show ' + self.image_name,
                parse_output=True,
            )
            image_id = cmd_output['id']

            # This means create a 1GB volume from the specified image, attach
            # it to the server at /dev/vdb and delete the volume when the
            # server is deleted.
            bdm_arg = (
                f'--block-device '
                f'device_name=vdb,uuid={image_id},'
                f'source_type=image,volume_size=1,'
                f'delete_on_termination=true,boot_index=1'
            )

        # create server with bdm type=image
        # NOTE(mriedem): This test is a bit unrealistic in that specifying the
        # same image in the block device as the --image option does not really
        # make sense, but we just want to make sure everything is processed
        # as expected where nova creates a volume from the image and attaches
        # that volume to the server.
        server_name = uuid.uuid4().hex
        server = self.openstack(
            'server create '
            + '--flavor '
            + self.flavor_name
            + ' '
            + '--image '
            + self.image_name
            + ' '
            + bdm_arg
            + ' '
            + self.network_arg
            + ' '
            + '--wait '
            + server_name,
            parse_output=True,
        )
        self.assertIsNotNone(server["id"])
        self.assertEqual(
            server_name,
            server['name'],
        )
        self.wait_for_status(server_name, 'ACTIVE')

        # check server volumes_attached, format is
        # {"volumes_attached": "id='2518bc76-bf0b-476e-ad6b-571973745bb5'",}
        cmd_output = self.openstack(
            'server show ' + server_name,
            parse_output=True,
        )
        volumes_attached = cmd_output['volumes_attached']
        self.assertIsNotNone(volumes_attached)
        attached_volume_id = volumes_attached[0]["id"]

        # check the volume that attached on server
        cmd_output = self.openstack(
            'volume show ' + attached_volume_id,
            parse_output=True,
        )
        attachments = cmd_output['attachments']
        self.assertEqual(
            1,
            len(attachments),
        )
        self.assertEqual(
            server['id'],
            attachments[0]['server_id'],
        )
        self.assertEqual(
            "in-use",
            cmd_output['status'],
        )
        # TODO(mriedem): If we can parse the volume_image_metadata field from
        # the volume show output we could assert the image_name is what we
        # specified. volume_image_metadata is something like this:
        # {u'container_format': u'bare', u'min_ram': u'0',
        # u'disk_format': u'qcow2', u'image_name': u'cirros-0.4.0-x86_64-disk',
        # u'image_id': u'05496c83-e2df-4c2f-9e48-453b6e49160d',
        # u'checksum': u'443b7623e27ecf03dc9e01ee93f67afe', u'min_disk': u'0',
        # u'size': u'12716032'}

        # delete server, then check the attached volume has been deleted
        self.openstack('server delete --wait ' + server_name)
        cmd_output = self.openstack(
            'volume list',
            parse_output=True,
        )
        target_volume = [
            each_volume
            for each_volume in cmd_output
            if each_volume['ID'] == attached_volume_id
        ]
        if target_volume:
            # check the attached volume is 'deleting' status
            self.assertEqual('deleting', target_volume[0]['Status'])
        else:
            # the attached volume had been deleted
            pass

    def test_server_boot_with_bdm_image(self):
        self._test_server_boot_with_bdm_image(use_legacy=False)

    # TODO(stephenfin): Remove when we drop support for the
    # '--block-device-mapping' option
    def test_server_boot_with_bdm_image_legacy(self):
        self._test_server_boot_with_bdm_image(use_legacy=True)

    def test_boot_from_volume(self):
        # Tests creating a server using --image and --boot-from-volume where
        # the compute service will create a root volume of the specified size
        # using the provided image, attach it as the root disk for the server
        # and not delete the volume when the server is deleted.
        server_name = uuid.uuid4().hex
        server = self.openstack(
            'server create '
            + '--flavor '
            + self.flavor_name
            + ' '
            + '--image '
            + self.image_name
            + ' '
            + '--boot-from-volume 1 '
            + self.network_arg  # create a 1GB volume from the image
            + ' '
            + '--wait '
            + server_name,
            parse_output=True,
        )
        self.assertIsNotNone(server["id"])
        self.assertEqual(
            server_name,
            server['name'],
        )
        self.wait_for_status(server_name, 'ACTIVE')

        # check server volumes_attached, format is
        # {"volumes_attached": "id='2518bc76-bf0b-476e-ad6b-571973745bb5'",}
        cmd_output = self.openstack(
            'server show ' + server_name,
            parse_output=True,
        )
        volumes_attached = cmd_output['volumes_attached']
        self.assertIsNotNone(volumes_attached)
        attached_volume_id = volumes_attached[0]["id"]
        for vol in volumes_attached:
            self.assertIsNotNone(vol['id'])
            # Don't leak the volume when the test exits.
            self.addCleanup(self.openstack, 'volume delete ' + vol['id'])

        # Since the server is volume-backed the GET /servers/{server_id}
        # response will have image='N/A (booted from volume)'.
        self.assertEqual(v2_server.IMAGE_STRING_FOR_BFV, cmd_output['image'])

        # check the volume that attached on server
        cmd_output = self.openstack(
            'volume show ' + volumes_attached[0]["id"],
            parse_output=True,
        )
        # The volume size should be what we specified on the command line.
        self.assertEqual(1, int(cmd_output['size']))
        attachments = cmd_output['attachments']
        self.assertEqual(
            1,
            len(attachments),
        )
        self.assertEqual(
            server['id'],
            attachments[0]['server_id'],
        )
        self.assertEqual(
            "in-use",
            cmd_output['status'],
        )
        # TODO(mriedem): If we can parse the volume_image_metadata field from
        # the volume show output we could assert the image_name is what we
        # specified. volume_image_metadata is something like this:
        # {u'container_format': u'bare', u'min_ram': u'0',
        # u'disk_format': u'qcow2', u'image_name': u'cirros-0.4.0-x86_64-disk',
        # u'image_id': u'05496c83-e2df-4c2f-9e48-453b6e49160d',
        # u'checksum': u'443b7623e27ecf03dc9e01ee93f67afe', u'min_disk': u'0',
        # u'size': u'12716032'}

        # delete server, then check the attached volume was not deleted
        self.openstack('server delete --wait ' + server_name)
        cmd_output = self.openstack(
            'volume show ' + attached_volume_id,
            parse_output=True,
        )
        # check the volume is in 'available' status
        self.assertEqual('available', cmd_output['status'])

    def test_server_create_with_none_network(self):
        """Test server create with none network option."""
        server_name = uuid.uuid4().hex
        server = self.openstack(
            # auto/none enable in nova micro version (v2.37+)
            '--os-compute-api-version 2.37 '
            + 'server create '
            + '--flavor '
            + self.flavor_name
            + ' '
            + '--image '
            + self.image_name
            + ' '
            + '--nic none '
            + server_name,
            parse_output=True,
        )
        self.assertIsNotNone(server["id"])
        self.addCleanup(self.openstack, 'server delete --wait ' + server_name)
        self.assertEqual(server_name, server['name'])
        self.wait_for_status(server_name, "ACTIVE")
        server = self.openstack(
            'server show ' + server_name,
            parse_output=True,
        )
        self.assertEqual({}, server['addresses'])

    def test_server_create_with_security_group(self):
        """Test server create with security group ID and name"""
        if not self.haz_network:
            # NOTE(dtroyer): As of Ocata release Nova forces nova-network to
            #                run in a cells v1 configuration.  Security group
            #                and network functions currently do not work in
            #                the gate jobs so we have to skip this.  It is
            #                known to work tested against a Mitaka nova-net
            #                DevStack without cells.
            self.skipTest("No Network service present")
        # Create two security group, use name and ID to create server
        sg_name1 = uuid.uuid4().hex
        security_group1 = self.openstack(
            'security group create ' + sg_name1,
            parse_output=True,
        )
        self.addCleanup(self.openstack, 'security group delete ' + sg_name1)
        sg_name2 = uuid.uuid4().hex
        security_group2 = self.openstack(
            'security group create ' + sg_name2,
            parse_output=True,
        )
        self.addCleanup(self.openstack, 'security group delete ' + sg_name2)

        server_name = uuid.uuid4().hex
        server = self.openstack(
            'server create '
            + '--flavor '
            + self.flavor_name
            + ' '
            + '--image '
            + self.image_name
            + ' '
            +
            # Security group id is integer in nova-network, convert to string
            '--security-group '
            + str(security_group1['id'])
            + ' '
            + '--security-group '
            + security_group2['name']
            + ' '
            + self.network_arg
            + ' '
            + server_name,
            parse_output=True,
        )
        self.addCleanup(self.openstack, 'server delete --wait ' + server_name)

        self.assertIsNotNone(server['id'])
        self.assertEqual(server_name, server['name'])
        sec_grp = ""
        for sec in server['security_groups']:
            sec_grp += sec['name']
        self.assertIn(str(security_group1['id']), sec_grp)
        self.assertIn(str(security_group2['id']), sec_grp)
        self.wait_for_status(server_name, 'ACTIVE')
        server = self.openstack(
            'server show ' + server_name,
            parse_output=True,
        )
        # check if security group exists in list
        sec_grp = ""
        for sec in server['security_groups']:
            sec_grp += sec['name']
        self.assertIn(sg_name1, sec_grp)
        self.assertIn(sg_name2, sec_grp)

    def test_server_create_with_empty_network_option_latest(self):
        """Test server create with empty network option in nova 2.latest."""
        server_name = uuid.uuid4().hex
        try:
            self.openstack(
                # auto/none enable in nova micro version (v2.37+)
                '--os-compute-api-version 2.37 '
                + 'server create '
                + '--flavor '
                + self.flavor_name
                + ' '
                + '--image '
                + self.image_name
                + ' '
                + server_name
            )
        except exceptions.CommandFailed as e:
            # If we got here, it shouldn't be because a nics value wasn't
            # provided to the server; it is likely due to something else in
            # the functional tests like there being multiple available
            # networks and the test didn't specify a specific network.
            self.assertNotIn(
                'nics are required after microversion 2.36', e.stderr
            )

    def test_server_add_remove_network(self):
        name = uuid.uuid4().hex
        cmd_output = self.openstack(
            'server create '
            + '--network private '
            + '--flavor '
            + self.flavor_name
            + ' '
            + '--image '
            + self.image_name
            + ' '
            + '--wait '
            + name,
            parse_output=True,
        )

        self.assertIsNotNone(cmd_output['id'])
        self.assertEqual(name, cmd_output['name'])
        self.addCleanup(self.openstack, 'server delete --wait ' + name)

        # add network and check 'public' is in server show
        self.openstack('server add network ' + name + ' public')

        wait_time = 0
        while wait_time < 60:
            cmd_output = self.openstack(
                'server show ' + name,
                parse_output=True,
            )
            if 'public' not in cmd_output['addresses']:
                # Hang out for a bit and try again
                print('retrying add network check')
                wait_time += 10
                time.sleep(10)
            else:
                break
        addresses = cmd_output['addresses']
        self.assertIn('public', addresses)

        # remove network and check 'public' is not in server show
        self.openstack('server remove network ' + name + ' public')

        wait_time = 0
        while wait_time < 60:
            cmd_output = self.openstack(
                'server show ' + name,
                parse_output=True,
            )
            if 'public' in cmd_output['addresses']:
                # Hang out for a bit and try again
                print('retrying remove network check')
                wait_time += 10
                time.sleep(10)
            else:
                break

        addresses = cmd_output['addresses']
        self.assertNotIn('public', addresses)

    def test_server_add_remove_port(self):
        name = uuid.uuid4().hex
        cmd_output = self.openstack(
            'server create '
            + '--network private '
            + '--flavor '
            + self.flavor_name
            + ' '
            + '--image '
            + self.image_name
            + ' '
            + '--wait '
            + name,
            parse_output=True,
        )

        self.assertIsNotNone(cmd_output['id'])
        self.assertEqual(name, cmd_output['name'])
        self.addCleanup(self.openstack, 'server delete --wait ' + name)

        # create port, record one of its ip address
        port_name = uuid.uuid4().hex

        cmd_output = self.openstack(
            'port list',
            parse_output=True,
        )
        self.assertNotIn(port_name, cmd_output)

        cmd_output = self.openstack(
            'port create ' + '--network private ' + port_name,
            parse_output=True,
        )
        self.assertIsNotNone(cmd_output['id'])
        ip_address = cmd_output['fixed_ips'][0]['ip_address']
        self.addCleanup(self.openstack, 'port delete ' + port_name)

        # add port to server, assert the ip address of the port appears
        self.openstack('server add port ' + name + ' ' + port_name)

        wait_time = 0
        while wait_time < 60:
            cmd_output = self.openstack(
                'server show ' + name,
                parse_output=True,
            )
            if ip_address not in cmd_output['addresses']['private']:
                # Hang out for a bit and try again
                print('retrying add port check')
                wait_time += 10
                time.sleep(10)
            else:
                break
        addresses = cmd_output['addresses']['private']
        self.assertIn(ip_address, addresses)

        # remove port, assert the ip address of the port doesn't appear
        self.openstack('server remove port ' + name + ' ' + port_name)

        wait_time = 0
        while wait_time < 60:
            cmd_output = self.openstack(
                'server show ' + name,
                parse_output=True,
            )
            if ip_address in cmd_output['addresses']['private']:
                # Hang out for a bit and try again
                print('retrying remove port check')
                wait_time += 10
                time.sleep(10)
            else:
                break
        addresses = cmd_output['addresses']['private']
        self.assertNotIn(ip_address, addresses)

    def test_server_add_fixed_ip(self):
        name = uuid.uuid4().hex
        cmd_output = self.openstack(
            'server create '
            + '--network private '
            + '--flavor '
            + self.flavor_name
            + ' '
            + '--image '
            + self.image_name
            + ' '
            + '--wait '
            + name,
            parse_output=True,
        )

        self.assertIsNotNone(cmd_output['id'])
        self.assertEqual(name, cmd_output['name'])
        self.addCleanup(self.openstack, 'server delete --wait ' + name)

        # create port, record its ip address to use in later call,
        # then delete - this is to figure out what should be a free ip
        # in the subnet
        port_name = uuid.uuid4().hex

        cmd_output = self.openstack(
            'port list',
            parse_output=True,
        )
        self.assertNotIn(port_name, cmd_output)

        cmd_output = self.openstack(
            'port create ' + '--network private ' + port_name,
            parse_output=True,
        )
        self.assertIsNotNone(cmd_output['id'])
        ip_address = cmd_output['fixed_ips'][0]['ip_address']
        self.openstack('port delete ' + port_name)

        # add fixed ip to server, assert the ip address appears
        self.openstack(
            'server add fixed ip --fixed-ip-address '
            + ip_address
            + ' '
            + name
            + ' private'
        )

        wait_time = 0
        while wait_time < 60:
            cmd_output = self.openstack(
                'server show ' + name,
                parse_output=True,
            )
            if ip_address not in cmd_output['addresses']['private']:
                # Hang out for a bit and try again
                print('retrying add fixed ip check')
                wait_time += 10
                time.sleep(10)
            else:
                break
        addresses = cmd_output['addresses']['private']
        self.assertIn(ip_address, addresses)

    def test_server_add_remove_security_group(self):
        name = uuid.uuid4().hex
        cmd_output = self.openstack(
            'server create '
            + '--network private '
            + '--flavor '
            + self.flavor_name
            + ' '
            + '--image '
            + self.image_name
            + ' '
            + '--wait '
            + name,
            parse_output=True,
        )

        self.assertIsNotNone(cmd_output['id'])
        self.assertEqual(name, cmd_output['name'])
        self.addCleanup(self.openstack, 'server delete --wait ' + name)

        # create security group
        security_group_name = uuid.uuid4().hex

        cmd_output = self.openstack(
            'security group list',
            parse_output=True,
        )
        self.assertNotIn(security_group_name, cmd_output)

        cmd_output = self.openstack(
            'security group create ' + security_group_name,
            parse_output=True,
        )
        self.assertIsNotNone(cmd_output['id'])
        self.addCleanup(
            self.openstack, 'security group delete ' + security_group_name
        )

        # add security group to server, assert the name of the security group
        # appears
        self.openstack(
            'server add security group ' + name + ' ' + security_group_name
        )

        wait_time = 0
        while wait_time < 60:
            cmd_output = self.openstack(
                'server show ' + name,
                parse_output=True,
            )
            if security_group_name not in [
                x['name'] for x in cmd_output['security_groups']
            ]:
                # Hang out for a bit and try again
                print('retrying add security group check')
                wait_time += 10
                time.sleep(10)
            else:
                break
        security_groups = [x['name'] for x in cmd_output['security_groups']]
        self.assertIn(security_group_name, security_groups)

        # remove security group, assert the name of the security group doesn't
        # appear
        self.openstack(
            'server remove security group ' + name + ' ' + security_group_name
        )

        wait_time = 0
        while wait_time < 60:
            cmd_output = self.openstack(
                'server show ' + name,
                parse_output=True,
            )
            if security_group_name not in [
                x['name'] for x in cmd_output['security_groups']
            ]:
                # Hang out for a bit and try again
                print('retrying remove security group check')
                wait_time += 10
                time.sleep(10)
            else:
                break
        security_groups = [x['name'] for x in cmd_output['security_groups']]
        self.assertNotIn(security_group_name, security_groups)

    def test_server_add_remove_volume(self):
        volume_wait_for = volume_common.BaseVolumeTests.wait_for_status

        server_name = uuid.uuid4().hex
        cmd_output = self.openstack(
            'server create '
            + '--network private '
            + '--flavor '
            + self.flavor_name
            + ' '
            + '--image '
            + self.image_name
            + ' '
            + '--wait '
            + server_name,
            parse_output=True,
        )

        self.assertIsNotNone(cmd_output['id'])
        self.assertEqual(server_name, cmd_output['name'])
        self.addCleanup(self.openstack, 'server delete --wait ' + server_name)
        server_id = cmd_output['id']

        volume_name = uuid.uuid4().hex
        cmd_output = self.openstack(
            'volume create ' + '--size 1 ' + volume_name,
            parse_output=True,
        )

        self.assertIsNotNone(cmd_output['id'])
        self.assertEqual(volume_name, cmd_output['name'])
        volume_wait_for('volume', volume_name, 'available')
        self.addCleanup(self.openstack, 'volume delete ' + volume_name)
        volume_id = cmd_output['id']

        cmd_output = self.openstack(
            'server add volume '
            + server_name
            + ' '
            + volume_name
            + ' '
            + '--tag bar',
            parse_output=True,
        )

        self.assertEqual(server_id, cmd_output['Server ID'])
        self.assertEqual(volume_id, cmd_output['Volume ID'])

        cmd_output = self.openstack(
            'server volume list ' + server_name,
            parse_output=True,
        )

        self.assertEqual(server_id, cmd_output[0]['Server ID'])
        self.assertEqual(volume_id, cmd_output[0]['Volume ID'])

        volume_wait_for('volume', volume_name, 'in-use')

        cmd_output = self.openstack(
            'server event list ' + server_name,
            parse_output=True,
        )
        self.assertEqual(2, len(cmd_output))
        self.assertIn('attach_volume', {x['Action'] for x in cmd_output})

        self.openstack(
            'server remove volume ' + server_name + ' ' + volume_name
        )
        volume_wait_for('volume', volume_name, 'available')

        cmd_output = self.openstack(
            'server event list ' + server_name,
            parse_output=True,
        )
        self.assertEqual(3, len(cmd_output))
        self.assertIn('detach_volume', {x['Action'] for x in cmd_output})

        raw_output = self.openstack('server volume list ' + server_name)
        self.assertEqual('\n', raw_output)

    def test_server_stop_start(self):
        """Test server stop, start"""
        server_name = uuid.uuid4().hex
        cmd_output = self.openstack(
            'server create '
            + '--network private '
            + '--flavor '
            + self.flavor_name
            + ' '
            + '--image '
            + self.image_name
            + ' '
            + '--wait '
            + server_name,
            parse_output=True,
        )

        self.assertIsNotNone(cmd_output['id'])
        self.assertEqual(server_name, cmd_output['name'])
        self.addCleanup(self.openstack, 'server delete --wait ' + server_name)
        server_id = cmd_output['id']

        cmd_output = self.openstack(
            'server stop ' + server_name,
        )
        self.assertEqual("", cmd_output)

        # This is our test that the request succeeded. If it doesn't transition
        # to SHUTOFF then it didn't work.
        self.wait_for_status(server_id, "SHUTOFF")

        cmd_output = self.openstack(
            'server start ' + server_name,
        )
        self.assertEqual("", cmd_output)

        # As above, this is our test that the request succeeded. If it doesn't
        # transition to ACTIVE then it didn't work.
        self.wait_for_status(server_id, "ACTIVE")

    def test_server_migration_list(self):
        # Verify that the command does not raise an exception when we list
        # migrations, including when we specify a query.
        self.openstack('server migration list')
        self.openstack('server migration list --limit 1')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
import logging
import os
import shlex
import subprocess

from tempest.lib.cli import output_parser
from tempest.lib import exceptions
import testtools

ADMIN_CLOUD = os.environ.get('OS_ADMIN_CLOUD', 'devstack-admin')
LOG = logging.getLogger(__name__)


def execute(cmd, *, fail_ok=False):
    """Executes specified command for the given action."""
    LOG.debug('Executing: %s', cmd)
    cmdlist = shlex.split(cmd)
    stdout = subprocess.PIPE
    stderr = subprocess.PIPE
    env = {
        k: v for k, v in os.environ.copy().items() if not k.startswith('OS_')
    }

    proc = subprocess.Popen(cmdlist, stdout=stdout, stderr=stderr, env=env)

    result_out_b, result_err = proc.communicate()
    result_out = result_out_b.decode('utf-8')
    LOG.debug('stdout: %s', result_out)
    LOG.debug('stderr: %s', result_err)

    if not fail_ok and proc.returncode != 0:
        raise exceptions.CommandFailed(
            proc.returncode,
            cmd,
            result_out,
            result_err,
        )

    return result_out


class TestCase(testtools.TestCase):
    @classmethod
    def openstack(
        cls,
        cmd,
        *,
        cloud=ADMIN_CLOUD,
        fail_ok=False,
        parse_output=False,
    ):
        """Executes openstackclient command for the given action

        :param cmd: A string representation of the command to execute.
        :param cloud: The cloud to execute against. This can be a string, empty
            string, or None. A string results in '--os-auth-type $cloud', an
            empty string results in the '--os-auth-type' option being
            omitted, and None results in '--os-auth-type none' for legacy
            reasons.
        :param fail_ok: If failure is permitted. If False (default), a command
            failure will result in `~tempest.lib.exceptions.CommandFailed`
            being raised.
        :param parse_output: If true, pass the '-f json' parameter and decode
            the output.
        :returns: The output from the command.
        :raises: `~tempest.lib.exceptions.CommandFailed` if the command failed
            and ``fail_ok`` was ``False``.
        """
        auth_args = []
        if cloud is None:
            # Execute command with no auth
            auth_args.append('--os-auth-type none')
        elif cloud != '':
            # Execute command with an explicit cloud specified
            auth_args.append(f'--os-cloud {cloud}')

        format_args = []
        if parse_output:
            format_args.append('-f json')

        output = execute(
            ' '.join(['openstack'] + auth_args + [cmd] + format_args),
            fail_ok=fail_ok,
        )

        if parse_output:
            try:
                return json.loads(output)
            except json.JSONDecodeError:
                print(f'failed to decode: {output}')
                raise
        else:
            return output

    @classmethod
    def is_service_enabled(cls, service, version=None):
        """Ask client cloud if service is available

        :param service: The service name or type. This should be either an
            exact match to what is in the catalog or a known official value or
            alias from service-types-authority
        :param version: Optional version. This should be a major version, e.g.
            '2.0'
        :returns: True if the service is enabled and optionally provides the
            specified API version, else False
        """
        ret = cls.openstack(
            f'versions show --service {service} -f value -c Version'
        ).splitlines()
        if version:
            return version in ret

        return bool(ret)

    @classmethod
    def is_extension_enabled(cls, alias, *, service='network'):
        """Ask client cloud if extension is enabled"""
        extensions = cls.openstack(
            f'extension list --{service}',
            parse_output=True,
        )
        return alias in [x['Alias'] for x in extensions]

    @classmethod
    def get_openstack_configuration_value(cls, configuration):
        opts = cls.get_opts([configuration])
        return cls.openstack('configuration show ' + opts)

    @classmethod
    def get_opts(cls, fields, output_format='value'):
        return ' -f {} {}'.format(
            output_format, ' '.join(['-c ' + it for it in fields])
        )

    @classmethod
    def assertOutput(cls, expected, actual):
        if expected != actual:
            raise Exception(expected + ' != ' + actual)

    @classmethod
    def assertInOutput(cls, expected, actual):
        if expected not in actual:
            raise Exception(expected + ' not in ' + actual)

    @classmethod
    def assertsOutputNotNone(cls, observed):
        if observed is None:
            raise Exception('No output observed')

    def assert_table_structure(self, items, field_names):
        """Verify that all items have keys listed in field_names."""
        for item in items:
            for field in field_names:
                self.assertIn(field, item)

    def assert_show_fields(self, show_output, field_names):
        """Verify that all items have keys listed in field_names."""

        # field_names = ['name', 'description']
        # show_output = [{'name': 'fc2b98d8faed4126b9e371eda045ade2'},
        #          {'description': 'description-821397086'}]
        # this next line creates a flattened list of all 'keys' (like 'name',
        # and 'description' out of the output
        all_headers = [item for sublist in show_output for item in sublist]
        for field_name in field_names:
            self.assertIn(field_name, all_headers)

    def parse_show_as_object(self, raw_output):
        """Return a dict with values parsed from cli output."""
        items = self.parse_show(raw_output)
        o = {}
        for item in items:
            o.update(item)
        return o

    def parse_show(self, raw_output):
        """Return list of dicts with item values parsed from cli output."""

        items = []
        table_ = output_parser.table(raw_output)
        for row in table_['values']:
            item = {}
            item[row[0]] = row[1]
            items.append(item)
        return items

    def parse_listing(self, raw_output):
        """Return list of dicts with basic item parsed from cli output."""
        return output_parser.listing(raw_output)



#   Copyright 2016 Easystack. All rights reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import http
import random
from unittest import mock
import uuid

from osc_lib import exceptions

from openstackclient.compute.v2 import agent
from openstackclient.tests.unit.compute.v2 import fakes as compute_fakes
from openstackclient.tests.unit import fakes
from openstackclient.tests.unit import utils as tests_utils


def _generate_fake_agent():
    return {
        'agent_id': random.randint(1, 1000),
        'os': 'agent-os-' + uuid.uuid4().hex,
        'architecture': 'agent-architecture',
        'version': '8.0',
        'url': 'http://127.0.0.1',
        'md5hash': 'agent-md5hash',
        'hypervisor': 'hypervisor',
    }


class TestAgentCreate(compute_fakes.TestComputev2):
    def setUp(self):
        super().setUp()

        self._agent = _generate_fake_agent()
        self.columns = (
            'agent_id',
            'architecture',
            'hypervisor',
            'md5hash',
            'os',
            'url',
            'version',
        )
        self.data = (
            self._agent['agent_id'],
            self._agent['architecture'],
            self._agent['hypervisor'],
            self._agent['md5hash'],
            self._agent['os'],
            self._agent['url'],
            self._agent['version'],
        )

        self.compute_client.post.return_value = fakes.FakeResponse(
            data={'agent': self._agent}
        )
        self.cmd = agent.CreateAgent(self.app, None)

    def test_agent_create(self):
        arglist = [
            self._agent['os'],
            self._agent['architecture'],
            self._agent['version'],
            self._agent['url'],
            self._agent['md5hash'],
            self._agent['hypervisor'],
        ]
        verifylist = [
            ('os', self._agent['os']),
            ('architecture', self._agent['architecture']),
            ('version', self._agent['version']),
            ('url', self._agent['url']),
            ('md5hash', self._agent['md5hash']),
            ('hypervisor', self._agent['hypervisor']),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.compute_client.post.assert_called_with(
            '/os-agents',
            json={
                'agent': {
                    'hypervisor': parsed_args.hypervisor,
                    'os': parsed_args.os,
                    'architecture': parsed_args.architecture,
                    'version': parsed_args.version,
                    'url': parsed_args.url,
                    'md5hash': parsed_args.md5hash,
                },
            },
            microversion='2.1',
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)


class TestAgentDelete(compute_fakes.TestComputev2):
    def setUp(self):
        super().setUp()

        self.compute_client.delete.return_value = fakes.FakeResponse(
            status_code=http.HTTPStatus.NO_CONTENT
        )

        self.cmd = agent.DeleteAgent(self.app, None)

    def test_delete_one_agent(self):
        arglist = ['123']
        verifylist = [
            ('id', ['123']),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.compute_client.delete.assert_called_once_with(
            '/os-agents/123',
            microversion='2.1',
        )
        self.assertIsNone(result)

    def test_delete_multiple_agents(self):
        arglist = ['1', '2', '3']
        verifylist = [
            ('id', ['1', '2', '3']),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        calls = [
            mock.call(f'/os-agents/{x}', microversion='2.1') for x in arglist
        ]
        self.compute_client.delete.assert_has_calls(calls)
        self.assertIsNone(result)

    def test_delete_multiple_agents_exception(self):
        arglist = ['1', '2', '999']
        verifylist = [
            ('id', ['1', '2', '999']),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.compute_client.delete.side_effect = [
            fakes.FakeResponse(status_code=http.HTTPStatus.NO_CONTENT),
            fakes.FakeResponse(status_code=http.HTTPStatus.NO_CONTENT),
            fakes.FakeResponse(status_code=http.HTTPStatus.NOT_FOUND),
        ]

        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        calls = [
            mock.call(f'/os-agents/{x}', microversion='2.1') for x in arglist
        ]
        self.compute_client.delete.assert_has_calls(calls)

    def test_agent_delete_no_input(self):
        arglist = []
        verifylist = None
        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )


class TestAgentList(compute_fakes.TestComputev2):
    def setUp(self):
        super().setUp()

        _agents = [_generate_fake_agent() for _ in range(3)]

        self.columns = (
            "Agent ID",
            "Hypervisor",
            "OS",
            "Architecture",
            "Version",
            "Md5Hash",
            "URL",
        )
        self.data = [
            (
                _agent['agent_id'],
                _agent['hypervisor'],
                _agent['os'],
                _agent['architecture'],
                _agent['version'],
                _agent['md5hash'],
                _agent['url'],
            )
            for _agent in _agents
        ]

        self.compute_client.get.return_value = fakes.FakeResponse(
            data={'agents': _agents},
        )
        self.cmd = agent.ListAgent(self.app, None)

    def test_agent_list(self):
        arglist = []
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))
        self.compute_client.get.assert_called_once_with(
            '/os-agents',
            microversion='2.1',
        )

    def test_agent_list_with_hypervisor(self):
        arglist = [
            '--hypervisor',
            'hypervisor',
        ]
        verifylist = [
            ('hypervisor', 'hypervisor'),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))
        self.compute_client.get.assert_called_once_with(
            '/os-agents?hypervisor=hypervisor',
            microversion='2.1',
        )


class TestAgentSet(compute_fakes.TestComputev2):
    def setUp(self):
        super().setUp()

        self.agent = _generate_fake_agent()
        self.compute_client.get.return_value = fakes.FakeResponse(
            data={'agents': [self.agent]},
        )
        self.compute_client.put.return_value = fakes.FakeResponse()

        self.cmd = agent.SetAgent(self.app, None)

    def test_agent_set_nothing(self):
        arglist = [
            str(self.agent['agent_id']),
        ]
        verifylist = [
            ('id', self.agent['agent_id']),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.compute_client.put.assert_called_once_with(
            f'/os-agents/{self.agent["agent_id"]}',
            json={
                'para': {
                    'version': self.agent['version'],
                    'url': self.agent['url'],
                    'md5hash': self.agent['md5hash'],
                },
            },
            microversion='2.1',
        )
        self.assertIsNone(result)

    def test_agent_set_version(self):
        arglist = [
            str(self.agent['agent_id']),
            '--agent-version',
            'new-version',
        ]

        verifylist = [
            ('id', self.agent['agent_id']),
            ('version', 'new-version'),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.compute_client.put.assert_called_once_with(
            f'/os-agents/{self.agent["agent_id"]}',
            json={
                'para': {
                    'version': parsed_args.version,
                    'url': self.agent['url'],
                    'md5hash': self.agent['md5hash'],
                },
            },
            microversion='2.1',
        )
        self.assertIsNone(result)

    def test_agent_set_url(self):
        arglist = [
            str(self.agent['agent_id']),
            '--url',
            'new-url',
        ]

        verifylist = [
            ('id', self.agent['agent_id']),
            ('url', 'new-url'),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.compute_client.put.assert_called_once_with(
            f'/os-agents/{self.agent["agent_id"]}',
            json={
                'para': {
                    'version': self.agent['version'],
                    'url': parsed_args.url,
                    'md5hash': self.agent['md5hash'],
                },
            },
            microversion='2.1',
        )
        self.assertIsNone(result)

    def test_agent_set_md5hash(self):
        arglist = [
            str(self.agent['agent_id']),
            '--md5hash',
            'new-md5hash',
        ]

        verifylist = [
            ('id', self.agent['agent_id']),
            ('md5hash', 'new-md5hash'),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.compute_client.put.assert_called_once_with(
            f'/os-agents/{self.agent["agent_id"]}',
            json={
                'para': {
                    'version': self.agent['version'],
                    'url': self.agent['url'],
                    'md5hash': parsed_args.md5hash,
                },
            },
            microversion='2.1',
        )
        self.assertIsNone(result)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os

from openstackclient.i18n import _


def bool_from_str(value, strict=False):
    true_strings = ('1', 't', 'true', 'on', 'y', 'yes')
    false_strings = ('0', 'f', 'false', 'off', 'n', 'no')

    if isinstance(value, bool):
        return value

    lowered = value.strip().lower()
    if lowered in true_strings:
        return True
    elif lowered in false_strings or not strict:
        return False

    msg = _(
        "Unrecognized value '%(value)s'; acceptable values are: %(valid)s"
    ) % {
        'value': value,
        'valid': ', '.join(
            f"'{s}'" for s in sorted(true_strings + false_strings)
        ),
    }
    raise ValueError(msg)


def boolenv(*vars, default=False):
    """Search for the first defined of possibly many bool-like env vars.

    Returns the first environment variable defined in vars, or returns the
    default.

    :param vars: Arbitrary strings to search for. Case sensitive.
    :param default: The default to return if no value found.
    :returns: A boolean corresponding to the value found, else the default if
        no value found.
    """
    for v in vars:
        value = os.environ.get(v, None)
        if value:
            return bool_from_str(value)
    return default



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest.lib.common.utils import data_utils

from openstackclient.tests.functional.identity.v3 import common


class EndpointTests(common.IdentityTests):
    def test_endpoint_create(self):
        self._create_dummy_endpoint(interface='public')
        self._create_dummy_endpoint(interface='admin')
        self._create_dummy_endpoint(interface='internal')

    def test_endpoint_delete(self):
        endpoint_id = self._create_dummy_endpoint(add_clean_up=False)
        raw_output = self.openstack(f'endpoint delete {endpoint_id}')
        self.assertEqual(0, len(raw_output))

    def test_endpoint_multi_delete(self):
        endpoint_1 = self._create_dummy_endpoint(add_clean_up=False)
        endpoint_2 = self._create_dummy_endpoint(add_clean_up=False)
        raw_output = self.openstack(
            f'endpoint delete {endpoint_1} {endpoint_2}'
        )
        self.assertEqual(0, len(raw_output))

    def test_endpoint_list(self):
        endpoint_id = self._create_dummy_endpoint()
        raw_output = self.openstack('endpoint list')
        self.assertIn(endpoint_id, raw_output)
        items = self.parse_listing(raw_output)
        self.assert_table_structure(items, self.ENDPOINT_LIST_HEADERS)

    def test_endpoint_list_filter(self):
        endpoint_id = self._create_dummy_endpoint(add_clean_up=False)
        project_id = self._create_dummy_project(add_clean_up=False)
        raw_output = self.openstack(
            f'endpoint add project {endpoint_id} {project_id}'
        )
        self.assertEqual(0, len(raw_output))
        raw_output = self.openstack(f'endpoint list --endpoint {endpoint_id}')
        self.assertIn(project_id, raw_output)
        items = self.parse_listing(raw_output)
        self.assert_table_structure(items, self.ENDPOINT_LIST_PROJECT_HEADERS)

        raw_output = self.openstack(f'endpoint list --project {project_id}')
        self.assertIn(endpoint_id, raw_output)
        items = self.parse_listing(raw_output)
        self.assert_table_structure(items, self.ENDPOINT_LIST_HEADERS)

    def test_endpoint_set(self):
        endpoint_id = self._create_dummy_endpoint()
        new_endpoint_url = data_utils.rand_url()
        raw_output = self.openstack(
            'endpoint set '
            '--interface {interface} '
            '--url {url} '
            '--disable '
            '{endpoint_id}'.format(
                interface='admin',
                url=new_endpoint_url,
                endpoint_id=endpoint_id,
            )
        )
        self.assertEqual(0, len(raw_output))
        raw_output = self.openstack(f'endpoint show {endpoint_id}')
        endpoint = self.parse_show_as_object(raw_output)
        self.assertEqual('admin', endpoint['interface'])
        self.assertEqual(new_endpoint_url, endpoint['url'])
        self.assertEqual('False', endpoint['enabled'])

    def test_endpoint_show(self):
        endpoint_id = self._create_dummy_endpoint()
        raw_output = self.openstack(f'endpoint show {endpoint_id}')
        items = self.parse_show(raw_output)
        self.assert_show_fields(items, self.ENDPOINT_FIELDS)

    def test_endpoint_add_remove_project(self):
        endpoint_id = self._create_dummy_endpoint(add_clean_up=False)
        project_id = self._create_dummy_project(add_clean_up=False)
        raw_output = self.openstack(
            f'endpoint add project {endpoint_id} {project_id}'
        )
        self.assertEqual(0, len(raw_output))

        raw_output = self.openstack(
            f'endpoint remove project {endpoint_id} {project_id}'
        )
        self.assertEqual(0, len(raw_output))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import unittest

import fixtures
from tempest.lib.common.utils import data_utils
from tempest.lib import exceptions as tempest_exceptions

from openstackclient.tests.functional import base

BASIC_LIST_HEADERS = ['ID', 'Name']


class IdentityTests(base.TestCase):
    """Functional tests for Identity commands."""

    USER_FIELDS = ['email', 'enabled', 'id', 'name', 'project_id', 'username']
    PROJECT_FIELDS = ['enabled', 'id', 'name', 'description']
    TOKEN_FIELDS = ['expires', 'id', 'project_id', 'user_id']
    ROLE_FIELDS = ['id', 'name', 'domain_id']
    SERVICE_FIELDS = ['id', 'enabled', 'name', 'type', 'description']
    ENDPOINT_FIELDS = [
        'id',
        'region',
        'service_id',
        'service_name',
        'service_type',
        'publicurl',
        'adminurl',
        'internalurl',
    ]

    EC2_CREDENTIALS_FIELDS = [
        'access',
        'project_id',
        'secret',
        'trust_id',
        'user_id',
    ]
    EC2_CREDENTIALS_LIST_HEADERS = [
        'Access',
        'Secret',
        'Project ID',
        'User ID',
    ]
    CATALOG_LIST_HEADERS = ['Name', 'Type', 'Endpoints']
    ENDPOINT_LIST_HEADERS = ['ID', 'Region', 'Service Name', 'Service Type']

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        # create dummy project
        cls.project_name = data_utils.rand_name('TestProject')
        cls.project_description = data_utils.rand_name('description')
        try:
            cls.openstack(
                '--os-identity-api-version 2 '
                'project create '
                f'--description {cls.project_description} '
                '--enable '
                f'{cls.project_name}'
            )
        except tempest_exceptions.CommandFailed:
            # Good chance this is due to Identity v2 admin not being enabled
            # TODO(dtroyer): Actually determine if Identity v2 admin is
            #                enabled in the target cloud.  Tuens out OSC
            #                doesn't make this easy as it should (yet).
            raise unittest.case.SkipTest('No Identity v2 admin endpoint?')

    @classmethod
    def tearDownClass(cls):
        try:
            cls.openstack(
                '--os-identity-api-version 2 '
                f'project delete {cls.project_name}'
            )
        finally:
            super().tearDownClass()

    def setUp(self):
        super().setUp()
        # prepare v2 env
        ver_fixture = fixtures.EnvironmentVariable(
            'OS_IDENTITY_API_VERSION', '2.0'
        )
        self.useFixture(ver_fixture)
        auth_url = os.environ.get('OS_AUTH_URL')
        if auth_url:
            auth_url_fixture = fixtures.EnvironmentVariable(
                'OS_AUTH_URL', auth_url.replace('v3', 'v2.0')
            )
            self.useFixture(auth_url_fixture)

    def _create_dummy_project(self, add_clean_up=True):
        project_name = data_utils.rand_name('TestProject')
        project_description = data_utils.rand_name('description')
        raw_output = self.openstack(
            'project create '
            f'--description {project_description} '
            f'--enable {project_name}'
        )
        project = self.parse_show_as_object(raw_output)
        if add_clean_up:
            self.addCleanup(
                self.openstack, 'project delete {}'.format(project['id'])
            )
        items = self.parse_show(raw_output)
        self.assert_show_fields(items, self.PROJECT_FIELDS)
        return project_name

    def _create_dummy_user(self, add_clean_up=True):
        username = data_utils.rand_name('TestUser')
        password = data_utils.rand_name('password')
        email = data_utils.rand_name() + '@example.com'
        raw_output = self.openstack(
            'user create '
            f'--project {self.project_name} '
            f'--password {password} '
            f'--email {email} '
            '--enable '
            f'{username}'
        )
        if add_clean_up:
            self.addCleanup(
                self.openstack,
                'user delete {}'.format(
                    self.parse_show_as_object(raw_output)['id']
                ),
            )
        items = self.parse_show(raw_output)
        self.assert_show_fields(items, self.USER_FIELDS)
        return username

    def _create_dummy_role(self, add_clean_up=True):
        role_name = data_utils.rand_name('TestRole')
        raw_output = self.openstack(f'role create {role_name}')
        role = self.parse_show_as_object(raw_output)
        if add_clean_up:
            self.addCleanup(
                self.openstack, 'role delete {}'.format(role['id'])
            )
        items = self.parse_show(raw_output)
        self.assert_show_fields(items, self.ROLE_FIELDS)
        self.assertEqual(role_name, role['name'])
        return role_name

    def _create_dummy_ec2_credentials(self, add_clean_up=True):
        raw_output = self.openstack('ec2 credentials create')
        ec2_credentials = self.parse_show_as_object(raw_output)
        access_key = ec2_credentials['access']
        if add_clean_up:
            self.addCleanup(
                self.openstack, f'ec2 credentials delete {access_key}'
            )
        items = self.parse_show(raw_output)
        self.assert_show_fields(items, self.EC2_CREDENTIALS_FIELDS)
        return access_key

    def _create_dummy_token(self, add_clean_up=True):
        raw_output = self.openstack('token issue')
        token = self.parse_show_as_object(raw_output)
        if add_clean_up:
            self.addCleanup(
                self.openstack, 'token revoke {}'.format(token['id'])
            )
        items = self.parse_show(raw_output)
        self.assert_show_fields(items, self.TOKEN_FIELDS)
        return token['id']

    def _create_dummy_service(self, add_clean_up=True):
        service_name = data_utils.rand_name('TestService')
        description = data_utils.rand_name('description')
        type_name = data_utils.rand_name('TestType')
        raw_output = self.openstack(
            'service create '
            f'--name {service_name} '
            f'--description {description} '
            f'{type_name}'
        )
        if add_clean_up:
            service = self.parse_show_as_object(raw_output)
            self.addCleanup(
                self.openstack, 'service delete {}'.format(service['id'])
            )
        items = self.parse_show(raw_output)
        self.assert_show_fields(items, self.SERVICE_FIELDS)
        return service_name

    def _create_dummy_endpoint(self, add_clean_up=True):
        region_id = data_utils.rand_name('TestRegion')
        service_name = self._create_dummy_service()
        public_url = data_utils.rand_url()
        admin_url = data_utils.rand_url()
        internal_url = data_utils.rand_url()
        raw_output = self.openstack(
            'endpoint create '
            f'--publicurl {public_url} '
            f'--adminurl {admin_url} '
            f'--internalurl {internal_url} '
            f'--region {region_id} '
            f'{service_name}'
        )
        endpoint = self.parse_show_as_object(raw_output)
        if add_clean_up:
            self.addCleanup(
                self.openstack, 'endpoint delete {}'.format(endpoint['id'])
            )
        items = self.parse_show(raw_output)
        self.assert_show_fields(items, self.ENDPOINT_FIELDS)
        return endpoint['id']



#   Copyright 2012-2013 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Image V2 Action Implementations"""

import logging

from osc_lib.cli import format_columns
from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _

_formatters = {
    'tags': format_columns.ListColumn,
}

LOG = logging.getLogger(__name__)


def _format_namespace(namespace):
    info = {}

    fields_to_show = [
        'created_at',
        'description',
        'display_name',
        'namespace',
        'owner',
        'protected',
        'schema',
        'updated_at',
        'visibility',
    ]

    namespace = namespace.to_dict(ignore_none=True, original_names=True)

    # split out the usual key and the properties which are top-level
    for key in namespace:
        if key in fields_to_show:
            info[key] = namespace.get(key)
        elif key == "resource_type_associations":
            info[key] = [
                resource_type['name'] for resource_type in namespace.get(key)
            ]
        elif key == 'properties':
            info['properties'] = list(namespace.get(key).keys())

    return info


class CreateMetadefNamespace(command.ShowOne):
    _description = _("Create a metadef namespace")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "namespace",
            metavar="<namespace>",
            help=_("New metadef namespace name"),
        )
        parser.add_argument(
            "--display-name",
            metavar="<display_name>",
            help=_("A user-friendly name for the namespace."),
        )
        parser.add_argument(
            "--description",
            metavar="<description>",
            help=_("A description of the namespace"),
        )
        visibility_group = parser.add_mutually_exclusive_group()
        visibility_group.add_argument(
            "--public",
            action="store_const",
            const="public",
            dest="visibility",
            help=_("Set namespace visibility 'public'"),
        )
        visibility_group.add_argument(
            "--private",
            action="store_const",
            const="private",
            dest="visibility",
            help=_("Set namespace visibility 'private'"),
        )
        protected_group = parser.add_mutually_exclusive_group()
        protected_group.add_argument(
            "--protected",
            action="store_const",
            const=True,
            dest="is_protected",
            help=_("Prevent metadef namespace from being deleted"),
        )
        protected_group.add_argument(
            "--unprotected",
            action="store_const",
            const=False,
            dest="is_protected",
            help=_("Allow metadef namespace to be deleted (default)"),
        )
        return parser

    def take_action(self, parsed_args):
        image_client = self.app.client_manager.image
        filter_keys = ['namespace', 'display_name', 'description']
        kwargs = {}

        for key in filter_keys:
            argument = getattr(parsed_args, key, None)
            if argument is not None:
                kwargs[key] = argument

        if parsed_args.is_protected is not None:
            kwargs['protected'] = parsed_args.is_protected

        if parsed_args.visibility is not None:
            kwargs['visibility'] = parsed_args.visibility

        data = image_client.create_metadef_namespace(**kwargs)
        info = _format_namespace(data)

        return zip(*sorted(info.items()))


class DeleteMetadefNamespace(command.Command):
    _description = _("Delete metadef namespace")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "namespace",
            metavar="<namespace>",
            nargs="+",
            help=_("Metadef namespace(s) to delete (name)"),
        )
        return parser

    def take_action(self, parsed_args):
        image_client = self.app.client_manager.image

        result = 0
        for ns in parsed_args.namespace:
            try:
                namespace = image_client.get_metadef_namespace(ns)
                image_client.delete_metadef_namespace(namespace.id)
            except Exception as e:
                result += 1
                LOG.error(
                    _(
                        "Failed to delete namespace with name or "
                        "ID '%(namespace)s': %(e)s"
                    ),
                    {'namespace': ns, 'e': e},
                )

        if result > 0:
            total = len(parsed_args.namespace)
            msg = _("%(result)s of %(total)s namespace failed to delete.") % {
                'result': result,
                'total': total,
            }
            raise exceptions.CommandError(msg)


class ListMetadefNamespace(command.Lister):
    _description = _("List metadef namespaces")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "--resource-types",
            metavar="<resource_types>",
            help=_("filter resource types"),
        )
        parser.add_argument(
            "--visibility",
            metavar="<visibility>",
            help=_("filter on visibility"),
        )
        return parser

    def take_action(self, parsed_args):
        image_client = self.app.client_manager.image
        filter_keys = ['resource_types', 'visibility']
        kwargs = {}
        for key in filter_keys:
            argument = getattr(parsed_args, key, None)
            if argument is not None:
                kwargs[key] = argument
        # List of namespace data received
        data = list(image_client.metadef_namespaces(**kwargs))
        columns = ['namespace']
        column_headers = columns
        return (
            column_headers,
            (
                utils.get_item_properties(
                    s,
                    columns,
                    formatters=_formatters,
                )
                for s in data
            ),
        )


class SetMetadefNamespace(command.Command):
    _description = _("Set metadef namespace properties")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "namespace",
            metavar="<namespace>",
            help=_("Metadef namespace to modify (name)"),
        )
        parser.add_argument(
            "--display-name",
            metavar="<display_name>",
            help=_("Set a user-friendly name for the namespace."),
        )
        parser.add_argument(
            "--description",
            metavar="<description>",
            help=_("Set the description of the namespace"),
        )
        visibility_group = parser.add_mutually_exclusive_group()
        visibility_group.add_argument(
            "--public",
            action="store_const",
            const="public",
            dest="visibility",
            help=_("Metadef namespace is accessible to the public"),
        )
        visibility_group.add_argument(
            "--private",
            action="store_const",
            const="private",
            dest="visibility",
            help=_(
                "Metadef namespace is inaccessible to the public (default)"
            ),
        )
        protected_group = parser.add_mutually_exclusive_group()
        protected_group.add_argument(
            "--protected",
            action="store_const",
            const=True,
            dest="is_protected",
            help=_("Prevent metadef namespace from being deleted"),
        )
        protected_group.add_argument(
            "--unprotected",
            action="store_const",
            const=False,
            dest="is_protected",
            help=_("Allow metadef namespace to be deleted (default)"),
        )
        return parser

    def take_action(self, parsed_args):
        image_client = self.app.client_manager.image

        namespace = parsed_args.namespace

        filter_keys = ['namespace', 'display_name', 'description']
        kwargs = {}

        for key in filter_keys:
            argument = getattr(parsed_args, key, None)
            if argument is not None:
                kwargs[key] = argument

        if parsed_args.is_protected is not None:
            kwargs['protected'] = parsed_args.is_protected

        if parsed_args.visibility is not None:
            kwargs['visibility'] = parsed_args.visibility

        image_client.update_metadef_namespace(namespace, **kwargs)


class ShowMetadefNamespace(command.ShowOne):
    _description = _("Show a metadef namespace")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "namespace",
            metavar="<namespace>",
            help=_("Metadef namespace to show (name)"),
        )
        return parser

    def take_action(self, parsed_args):
        image_client = self.app.client_manager.image

        namespace = parsed_args.namespace

        data = image_client.get_metadef_namespace(namespace)
        info = _format_namespace(data)

        return zip(*sorted(info.items()))



#   Copyright 2018 SUSE Linux GmbH
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Identity v3 Application Credential action implementations"""

import datetime
import json
import logging
import typing as ty
import uuid

from cliff import columns as cliff_columns
from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _
from openstackclient.identity import common

LOG = logging.getLogger(__name__)


class RolesColumn(cliff_columns.FormattableColumn[ty.Any]):
    """Generate a formatted string of role names."""

    def human_readable(self):
        return utils.format_list(list(r['name'] for r in self._value))


def _format_application_credential(
    application_credential, *, include_secret=False
):
    column_headers: tuple[str, ...] = (
        'ID',
        'Name',
        'Description',
        'Project ID',
        'Roles',
        'Unrestricted',
        'Access Rules',
        'Expires At',
    )
    columns: tuple[str, ...] = (
        'id',
        'name',
        'description',
        'project_id',
        'roles',
        'unrestricted',
        'access_rules',
        'expires_at',
    )
    if include_secret:
        column_headers += ('Secret',)
        columns += ('secret',)

    return (
        column_headers,
        utils.get_item_properties(
            application_credential, columns, formatters={'roles': RolesColumn}
        ),
    )


def _format_application_credentials(application_credentials):
    column_headers = (
        'ID',
        'Name',
        'Description',
        'Project ID',
        'Roles',
        'Unrestricted',
        'Access Rules',
        'Expires At',
    )
    columns = (
        'id',
        'name',
        'description',
        'project_id',
        'roles',
        'unrestricted',
        'access_rules',
        'expires_at',
    )

    return (
        column_headers,
        (
            utils.get_item_properties(
                x, columns, formatters={'roles': RolesColumn}
            )
            for x in application_credentials
        ),
    )


# TODO(stephenfin): Move this to osc_lib since it's useful elsewhere
def is_uuid_like(value) -> bool:
    """Returns validation of a value as a UUID.

    :param val: Value to verify
    :type val: string
    :returns: bool
    """
    try:
        formatted_value = (
            value.replace('urn:', '')
            .replace('uuid:', '')
            .strip('{}')
            .replace('-', '')
            .lower()
        )
        return str(uuid.UUID(value)).replace('-', '') == formatted_value
    except (TypeError, ValueError, AttributeError):
        return False


class CreateApplicationCredential(command.ShowOne):
    _description = _("Create new application credential")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'name',
            metavar='<name>',
            help=_('Name of the application credential'),
        )
        parser.add_argument(
            '--secret',
            metavar='<secret>',
            help=_(
                'Secret to use for authentication (if not provided, one'
                ' will be generated)'
            ),
        )
        parser.add_argument(
            '--role',
            metavar='<role>',
            dest='roles',
            action='append',
            default=[],
            help=_(
                'Roles to authorize (name or ID) (repeat option to set'
                ' multiple values)'
            ),
        )
        parser.add_argument(
            '--expiration',
            metavar='<expiration>',
            help=_(
                'Sets an expiration date for the application credential,'
                ' format of YYYY-mm-ddTHH:MM:SS (if not provided, the'
                ' application credential will not expire)'
            ),
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_('Application credential description'),
        )
        parser.add_argument(
            '--unrestricted',
            action="store_true",
            help=_(
                'Enable application credential to create and delete other'
                ' application credentials and trusts (this is potentially'
                ' dangerous behavior and is disabled by default)'
            ),
        )
        parser.add_argument(
            '--restricted',
            action="store_true",
            help=_(
                'Prohibit application credential from creating and deleting'
                ' other application credentials and trusts (this is the'
                ' default behavior)'
            ),
        )
        parser.add_argument(
            '--access-rules',
            metavar='<access-rules>',
            help=_(
                'Either a string or file path containing a JSON-formatted '
                'list of access rules, each containing a request method, '
                'path, and service, for example '
                '\'[{"method": "GET", '
                '"path": "/v2.1/servers", '
                '"service": "compute"}]\''
            ),
        )
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.sdk_connection.identity
        conn = self.app.client_manager.sdk_connection
        auth = conn.config.get_auth()
        if auth is None:
            # this will never happen
            raise exceptions.CommandError('invalid authentication info')

        user_id = auth.get_user_id(conn.identity)

        role_ids = []
        for role in parsed_args.roles:
            if is_uuid_like(role):
                role_ids.append({'id': role})
            else:
                role_ids.append({'name': role})

        expires_at = None
        if parsed_args.expiration:
            expires_at = datetime.datetime.strptime(
                parsed_args.expiration, '%Y-%m-%dT%H:%M:%S'
            )

        if parsed_args.restricted:
            unrestricted = False
        else:
            unrestricted = parsed_args.unrestricted

        if parsed_args.access_rules:
            try:
                access_rules = json.loads(parsed_args.access_rules)
            except ValueError:
                try:
                    with open(parsed_args.access_rules) as f:
                        access_rules = json.load(f)
                except OSError:
                    msg = _(
                        "Access rules is not valid JSON string or file does"
                        " not exist."
                    )
                    raise exceptions.CommandError(msg)
        else:
            access_rules = []

        application_credential = identity_client.create_application_credential(
            user_id,
            parsed_args.name,
            roles=role_ids,
            expires_at=expires_at,
            description=parsed_args.description,
            secret=parsed_args.secret,
            unrestricted=unrestricted,
            access_rules=access_rules,
        )

        return _format_application_credential(
            application_credential, include_secret=True
        )


class DeleteApplicationCredential(command.Command):
    _description = _("Delete application credentials(s)")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'application_credential',
            metavar='<application-credential>',
            nargs="+",
            help=_('Application credentials(s) to delete (name or ID)'),
        )
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.sdk_connection.identity
        conn = self.app.client_manager.sdk_connection
        auth = conn.config.get_auth()
        if auth is None:
            # this will never happen
            raise exceptions.CommandError('invalid authentication info')

        user_id = auth.get_user_id(conn.identity)

        errors = 0
        for ac in parsed_args.application_credential:
            try:
                app_cred = identity_client.find_application_credential(
                    user_id, ac, ignore_missing=False
                )
                identity_client.delete_application_credential(
                    user_id, app_cred.id
                )
            except Exception as e:
                errors += 1
                LOG.error(
                    _(
                        "Failed to delete application credential with "
                        "name or ID '%(ac)s': %(e)s"
                    ),
                    {'ac': ac, 'e': e},
                )

        if errors > 0:
            total = len(parsed_args.application_credential)
            msg = _(
                "%(errors)s of %(total)s application credentials failed "
                "to delete."
            ) % {'errors': errors, 'total': total}
            raise exceptions.CommandError(msg)

        return None


class ListApplicationCredential(command.Lister):
    _description = _("List application credentials")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--user',
            metavar='<user>',
            help=_('User whose application credentials to list (name or ID)'),
        )
        common.add_user_domain_option_to_parser(parser)
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.sdk_connection.identity
        if parsed_args.user:
            user_id = common.find_user_id_sdk(
                identity_client, parsed_args.user, parsed_args.user_domain
            )
        else:
            conn = self.app.client_manager.sdk_connection
            auth = conn.config.get_auth()
            if auth is None:
                # this will never happen
                raise exceptions.CommandError('invalid authentication info')
            user_id = auth.get_user_id(conn.identity)

        application_credentials = identity_client.application_credentials(
            user=user_id
        )

        return _format_application_credentials(application_credentials)


class ShowApplicationCredential(command.ShowOne):
    _description = _("Display application credential details")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'application_credential',
            metavar='<application-credential>',
            help=_('Application credential to display (name or ID)'),
        )
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.sdk_connection.identity
        conn = self.app.client_manager.sdk_connection
        auth = conn.config.get_auth()
        if auth is None:
            # this will never happen
            raise exceptions.CommandError('invalid authentication info')
        user_id = auth.get_user_id(conn.identity)

        application_credential = identity_client.find_application_credential(
            user_id, parsed_args.application_credential, ignore_missing=False
        )

        return _format_application_credential(application_credential)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from osc_lib.cli import format_columns

from openstackclient import command
from openstackclient.i18n import _


class ImportInfo(command.ShowOne):
    _description = _("Show available import methods")

    def take_action(self, parsed_args):
        image_client = self.app.client_manager.image

        import_info = image_client.get_import_info()
        import_methods = import_info.import_methods or {}
        return (
            ('import-methods',),
            (format_columns.ListColumn(import_methods.get('value', [])),),
        )
